/*
 * (SCCP*)
 *
 * An implementation of Skinny Client Control Protocol (SCCP)
 *
 * Sergio Chersovani (mlists@c-net.it)
 *
 * Reworked, but based on chan_sccp code.
 * The original chan_sccp driver that was made by Zozo which itself was derived from the chan_skinny driver.
 * Modified by Jan Czmok and Julien Goodwin
 *
 * This program is free software and may be modified and
 * distributed under the terms of the GNU Public License.
 */


#include "config.h"
#include "chan_sccp.h"
#include "sccp_actions.h"
#include "sccp_utils.h"
#include "sccp_device.h"
#include "sccp_channel.h"
#include "sccp_cli.h"
#include "sccp_line.h"
#include "sccp_socket.h"
#include "sccp_pbx.h"
#include "sccp_indicate.h"
#include <ctype.h>
#include <unistd.h>
#include <asterisk/pbx.h>
#include <asterisk/callerid.h>
#include <asterisk/utils.h>
#include <asterisk/causes.h>


static pthread_t socket_thread;

/* ---------------------------------------------------- */

#ifdef CS_AST_HAS_TECH_PVT
struct ast_channel *sccp_request(const char *type, int format, void *data, int *cause) {
#else
struct ast_channel *sccp_request(char *type, int format, void *data) {
#endif

  // type = SCCP
  // format = <codec>
  // data == line to call
  // sccp_pvt_t * = NULL;
  sccp_line_t * l = NULL;
  sccp_channel_t * c = NULL;
  /* sccp_intercom_t * i = NULL; not yet implemented XXXNEW*/

  if (!data) {
	ast_log(LOG_NOTICE, "Attempt to call SCCP/ failed\n");
#ifdef CS_AST_HAS_TECH_PVT
	*cause = AST_CAUSE_REQUESTED_CHAN_UNAVAIL;
#endif
	return NULL;
  }

  sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Asterisk asked to create a channel type(%s) format (%d) data (%s)\n", type, format, (char*)data);

  l = sccp_line_find_byname((char*)data);

  if (!l) {
		sccp_log(1)(VERBOSE_PREFIX_3 "SCCP/%s does not exist!\n", (char*)data);
#ifdef CS_AST_HAS_TECH_PVT
		*cause = AST_CAUSE_REQUESTED_CHAN_UNAVAIL;
#endif
		return NULL;
	}

  if (l) {
	if (!l->device) {
		sccp_log(10)(VERBOSE_PREFIX_3 "SCCP/%s isn't currently registered anywhere.\n", l->name);
#ifdef CS_AST_HAS_TECH_PVT
		*cause = AST_CAUSE_REQUESTED_CHAN_UNAVAIL;
#endif
		return NULL;
	}
  }

	// Allocate a new SCCP channel.
	/* on multiline phone we set the line when answering or switching lines */
	/* XXX sccp_dev_set_activeline(l->device,l); */
	 c = sccp_channel_allocate(l);
	 if (!c) {
#ifdef CS_AST_HAS_TECH_PVT
		*cause = AST_CAUSE_REQUESTED_CHAN_UNAVAIL;
#endif
	 	return NULL;
	 }

  if (!sccp_pbx_channel_allocate(c)) {
#ifdef CS_AST_HAS_TECH_PVT
		*cause = AST_CAUSE_REQUESTED_CHAN_UNAVAIL;
#endif
	 	return NULL;
	 }

  return c->owner;
}

/*
int sccp_devicestate(void *data) {
  char *dest = (char *) data;
  sccp_line_t * l =  NULL;
  int res = AST_DEVICE_UNKNOWN;

  l = sccp_line_find_byname((char*)data);
  if (!l)
	res = AST_DEVICE_INVALID;
  else if (!l->channelCount)
	res = AST_DEVICE_NOT_INUSE;
  // in use... what type? let's see
// XXX old CVS HEAD does not have AST_DEVICE_RINGING
// we have to check for it in the Makefile
//	else if (l->activeChannel) {
//	if (l->activeChannel->state == SCCP_CHANNELSTATE_CONNECTED)
//		res = AST_DEVICE_INUSE;
//	else if (l->activeChannel->state == SCCP_CHANNELSTATE_CONNECTED)
//		res = AST_DEVICE_RINGING;
//  }


  else
		res = AST_DEVICE_INUSE;

  if (l && l->device) {
  	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Asterisk asked for the state (%d) of the line %s\n", l->device->id, res, l->name);
  } else {
  	sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Asterisk is checking the state of the line %s\n", dest);
  }

  return res;
}
*/

sccp_hint_t * sccp_hint_make(sccp_device_t *d, uint8_t instance) {
	sccp_hint_t *h = NULL;
	if (!d)
		return NULL;
	h = malloc(sizeof(sccp_hint_t));

	if (!h)
		return NULL;
	h->device = d;
	h->instance = instance;
	return h;
}

void sccp_hint_notify(sccp_channel_t *c) {
	sccp_hint_t *h;
	sccp_line_t *l = c->line;
	sccp_device_t *d;
	sccp_moo_t * r;

	h = l->hints;
	if (!h)
		return;
	d = h->device;
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: HINT notify the state of the line %s \n", d->id, l->name);

	while (h) {
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: HINT notify state %s (%d) of the channel %s-%d \n", d->id, sccp_callstate2str(c->callstate), c->callstate, l->name, c->callid);
		if (c->callstate == SKINNY_CALLSTATE_CONNECTED || c->callstate == SKINNY_CALLSTATE_OFFHOOK || c->callstate == SKINNY_CALLSTATE_RINGIN) {
			if (c->callstate == SKINNY_CALLSTATE_RINGIN) {
				sccp_dev_set_lamp(d, h->instance, SKINNY_LAMP_BLINK);
			} else {
				sccp_dev_set_lamp(d, h->instance, SKINNY_LAMP_ON);
			}
			sccp_channel_set_callstate1(d, h->instance, c->callid, SKINNY_CALLSTATE_CALLREMOTEMULTILINE);
			REQ(r, CallInfoMessage);
			sccp_copy_string(r->msg.CallInfoMessage.callingPartyName, c->callingPartyName, sizeof(r->msg.CallInfoMessage.callingPartyName));
			sccp_copy_string(r->msg.CallInfoMessage.callingParty, c->callingPartyNumber, sizeof(r->msg.CallInfoMessage.callingParty));

			sccp_copy_string(r->msg.CallInfoMessage.calledPartyName, c->calledPartyName, sizeof(r->msg.CallInfoMessage.calledPartyName));
			sccp_copy_string(r->msg.CallInfoMessage.calledParty, c->calledPartyNumber, sizeof(r->msg.CallInfoMessage.calledParty));

			r->msg.CallInfoMessage.lel_lineId   = htolel(h->instance);
			r->msg.CallInfoMessage.lel_callRef  = htolel(c->callid);
			r->msg.CallInfoMessage.lel_callType = htolel(c->calltype);
			sccp_dev_send(d, r);
//			sccp_dev_sendmsg(d, DeactivateCallPlaneMessage);
			sccp_dev_set_keyset(d, h->instance, c->callid, KEYMODE_MYST);
		} else if (c->callstate == SKINNY_CALLSTATE_ONHOOK) {
			sccp_dev_set_lamp(d, h->instance, SKINNY_LAMP_ON);
			sccp_channel_set_callstate1(d, h->instance, c->callid, SKINNY_CALLSTATE_ONHOOK);
		}
		h = h->next;
	}
}

/* asterisk hint wrapper */
int sccp_hint_state(char *context, char* exten, int state, void *data) {
	sccp_hint_t * h = data;
	sccp_device_t *d;
	sccp_moo_t * r;
	if (state == -1 || !h || !h->device)
		return 0;

	d = h->device;

	switch(state) {
		case AST_EXTENSION_NOT_INUSE:
			sccp_dev_set_lamp(d, h->instance, SKINNY_LAMP_OFF);
			sccp_channel_set_callstate1(d, h->instance, 0, SKINNY_CALLSTATE_ONHOOK);
			return 0;
		case AST_EXTENSION_INUSE:
		case AST_EXTENSION_BUSY:
			sccp_dev_set_lamp(d, h->instance, SKINNY_LAMP_ON);
			sccp_channel_set_callstate1(d, h->instance, 0, SKINNY_CALLSTATE_CALLREMOTEMULTILINE);
			break;
		case AST_EXTENSION_UNAVAILABLE:
			sccp_dev_set_lamp(d, h->instance, SKINNY_LAMP_FLASH);
			sccp_channel_set_callstate1(d, h->instance, 0, SKINNY_CALLSTATE_CALLREMOTEMULTILINE);
			break;
		default:
			sccp_dev_set_lamp(d, h->instance, SKINNY_LAMP_FLASH);
			sccp_channel_set_callstate1(d, h->instance, 0, SKINNY_CALLSTATE_CALLREMOTEMULTILINE);
	}
	REQ(r, CallInfoMessage);
	sccp_copy_string(r->msg.CallInfoMessage.callingPartyName, "Unknown Number", sizeof(r->msg.CallInfoMessage.callingPartyName));
	sccp_copy_string(r->msg.CallInfoMessage.callingParty, "", sizeof(r->msg.CallInfoMessage.callingParty));
	sccp_copy_string(r->msg.CallInfoMessage.calledPartyName, "Unknown Number", sizeof(r->msg.CallInfoMessage.calledPartyName));
	sccp_copy_string(r->msg.CallInfoMessage.calledParty, "", sizeof(r->msg.CallInfoMessage.calledParty));

	r->msg.CallInfoMessage.lel_lineId   = htolel(h->instance);
	r->msg.CallInfoMessage.lel_callRef  = htolel(0);
	r->msg.CallInfoMessage.lel_callType = htolel(SKINNY_CALLTYPE_OUTBOUND);
	sccp_dev_send(d, r);
//	sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: HINT state %d of %s on line %s\n", state, hint, l->name);
	return 0;
}

int sccp_handle_message(sccp_moo_t * r, sccp_session_t * s) {
  uint32_t  mid = letohl(r->lel_messageId);
  s->lastKeepAlive = time(0); /* always update keepalive */

  if ( (!s->device) && (mid != RegisterMessage && mid != UnregisterMessage && mid != AlarmMessage && mid != KeepAliveMessage && mid != IpPortMessage)) {
	ast_log(LOG_WARNING, "Client sent %s without first registering.\n", sccpmsg2str(mid));
	free(r);
	return 0;
  }

  if (mid != KeepAliveMessage) {
	if (s && s->device) {
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: >> Got message %s\n", s->device->id, sccpmsg2str(mid));
	} else {
		sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: >> Got message %s\n", sccpmsg2str(mid));
	}
  }

  switch (mid) {

  case AlarmMessage:
	sccp_handle_alarm(s, r);
	break;
  case RegisterMessage:
	sccp_handle_register(s, r);
	break;
  case UnregisterMessage:
	sccp_handle_unregister(s, r);
	break;
  case VersionReqMessage:
	sccp_handle_version(s, r);
	break;
  case CapabilitiesResMessage:
	sccp_handle_capabilities_res(s, r);
	break;
  case ButtonTemplateReqMessage:
	sccp_handle_button_template_req(s, r);
	break;
  case SoftKeyTemplateReqMessage:
	sccp_handle_soft_key_template_req(s, r);
	break;
  case SoftKeySetReqMessage:
	sccp_handle_soft_key_set_req(s, r);
	break;
  case LineStatReqMessage:
	sccp_handle_line_number(s, r);
	break;
  case SpeedDialStatReqMessage:
	sccp_handle_speed_dial_stat_req(s, r);
	break;
  case StimulusMessage:
	sccp_handle_stimulus(s, r);
	break;
  case OffHookMessage:
	sccp_handle_offhook(s, r);
	break;
  case OnHookMessage:
	sccp_handle_onhook(s, r);
	break;
  case HeadsetStatusMessage:
	sccp_handle_headset(s, r);
	break;
  case TimeDateReqMessage:
	sccp_handle_time_date_req(s, r);
	break;
  case KeypadButtonMessage:
	sccp_handle_keypad_button(s, r);
	break;
  case SoftKeyEventMessage:
	sccp_handle_soft_key_event(s, r);
	break;
  case KeepAliveMessage:
	sccp_session_sendmsg(s, KeepAliveAckMessage);
	sccp_dev_check_mwi(s->device);
	break;
  case IpPortMessage:
  	/* obsolete message */
	s->rtpPort = letohs(r->msg.IpPortMessage.les_rtpMediaPort);
	break;
  case OpenReceiveChannelAck:
	sccp_handle_open_receive_channel_ack(s, r);
	break;
  case ConnectionStatisticsRes:
	sccp_handle_ConnectionStatistics(s,r);
	break;
  case ServerReqMessage:
	sccp_handle_ServerResMessage(s,r);
	break;
  case ConfigStatReqMessage:
	sccp_handle_ConfigStatMessage(s,r);
	break;
  case EnblocCallMessage:
	sccp_handle_EnblocCallMessage(s,r);
	break;
  case RegisterAvailableLinesMessage:
	if (s->device)
		sccp_dev_set_registered(s->device, SKINNY_DEVICE_RS_OK);
	break;
  case ForwardStatReqMessage:
	sccp_handle_forward_stat_req(s,r);
	break;
  default:
	if (GLOB(debug))
		ast_log(LOG_WARNING, "Unhandled SCCP Message: %d - %s\n", mid, sccpmsg2str(mid));
  }

  free(r);
  return 1;
}

static sccp_line_t * build_line(void) {
	sccp_line_t * l = malloc(sizeof(sccp_line_t));
	if (!l) {
		sccp_log(0)(VERBOSE_PREFIX_3 "Unable to allocate memory for a line\n");
		return NULL;
	}
	memset(l, 0, sizeof(sccp_line_t));
	ast_mutex_init(&l->lock);
	l->instance = -1;
	l->incominglimit = 3; /* default value */
	l->echocancel = 1; /* default value */
	l->silencesuppression = 0; /* default value */
	l->rtptos = 184; /* default value */

	sccp_copy_string(l->context, GLOB(default_context), sizeof(l->context));

  return l;
}

static sccp_device_t * build_device(void) {
	sccp_device_t * d = malloc(sizeof(sccp_device_t));
	if (!d) {
		sccp_log(0)(VERBOSE_PREFIX_3 "Unable to allocate memory for a device\n");
		return NULL;
	}
	memset(d, 0, sizeof(sccp_device_t));
	ast_mutex_init(&d->lock);

	d->tz_offset = 0;
	d->capability = GLOB(global_capability);
	d->codecs = GLOB(global_codecs);
	d->transfer = 1;
	d->state = SCCP_DEVICESTATE_ONHOOK;
	d->ringermode = SKINNY_STATION_RINGOFF;

#ifdef CS_SCCP_PARK
	d->park = 1;
#else
	d->park = 0;
#endif

  return d;
}

static int reload_config(void) {
	struct ast_config	*cfg;
	struct ast_variable	*v;
	int					oldport	= ntohs(GLOB(bindaddr.sin_port));
	int					on		= 1;
	int					tos		= IPTOS_LOWDELAY;
	char				pref_buf[128];
	char				iabuf[INET_ADDRSTRLEN];
	struct ast_hostent	ahp;
	struct hostent		*hp;
	sccp_device_t		*d;
	sccp_line_t			*l;
	sccp_speed_t		*k = NULL, *k_last = NULL;
	char 				*splitter, *k_exten = NULL, *k_name = NULL, *k_hint = NULL;
	char k_speed[256];
	uint8_t 			speeddial_index = 1;
	int firstdigittimeout = 0;
	int digittimeout = 0;

	memset(&GLOB(global_codecs), 0, sizeof(GLOB(global_codecs)));
	memset(&GLOB(bindaddr), 0, sizeof(GLOB(bindaddr)));

	cfg = ast_config_load("sccp.conf");
	if (!cfg) {
		ast_log(LOG_WARNING, "Unable to load config file sccp.conf, SCCP disabled\n");
		return 0;
	}
	/* read the general section */
	v = ast_variable_browse(cfg, "general");
	if (!v) {
		ast_log(LOG_WARNING, "Missing [general] section, SCCP disabled\n");
		return 0;
	}

	while (v) {
		if (!strcasecmp(v->name, "bindaddr")) {
			if (!(hp = ast_gethostbyname(v->value, &ahp))) {
				ast_log(LOG_WARNING, "Invalid address: %s. SCCP disabled\n", v->value);
				return 0;
			} else {
				memcpy(&GLOB(bindaddr.sin_addr), hp->h_addr, sizeof(GLOB(bindaddr.sin_addr)));
			}
		} else if (!strcasecmp(v->name, "keepalive")) {
			GLOB(keepalive) = atoi(v->value);
		} else if (!strcasecmp(v->name, "context")) {
			sccp_copy_string(GLOB(default_context), v->value, sizeof(GLOB(default_context)));
		} else if (!strcasecmp(v->name, "dateformat")) {
			sccp_copy_string (GLOB(date_format), v->value, sizeof(GLOB(date_format)));
		} else if (!strcasecmp(v->name, "port")) {
			if (sscanf(v->value, "%i", &GLOB(ourport)) == 1) {
				GLOB(bindaddr.sin_port) = htons(GLOB(ourport));
			} else {
				ast_log(LOG_WARNING, "Invalid port number '%s' at line %d of SCCP.CONF\n", v->value, v->lineno);
			}
		} else if (!strcasecmp(v->name, "firstdigittimeout")) {
			if (sscanf(v->value, "%i", &firstdigittimeout) == 1) {
				if (firstdigittimeout > 0 && firstdigittimeout < 255)
					GLOB(firstdigittimeout) = firstdigittimeout;
			} else {
				ast_log(LOG_WARNING, "Invalid firstdigittimeout number '%s' at line %d of SCCP.CONF\n", v->value, v->lineno);
			}
		} else if (!strcasecmp(v->name, "digittimeout")) {
			if (sscanf(v->value, "%i", &digittimeout) == 1) {
				if (digittimeout > 0 && digittimeout < 255)
					GLOB(digittimeout) = digittimeout;
			} else {
				ast_log(LOG_WARNING, "Invalid firstdigittimeout number '%s' at line %d of SCCP.CONF\n", v->value, v->lineno);
			}
		} else if (!strcasecmp(v->name, "debug")) {
			GLOB(debug) = atoi(v->value);
		} else if (!strcasecmp(v->name, "allow")) {
			ast_parse_allow_disallow(&GLOB(global_codecs), &GLOB(global_capability), v->value, 1);
		} else if (!strcasecmp(v->name, "disallow")) {
			ast_parse_allow_disallow(&GLOB(global_codecs), &GLOB(global_capability), v->value, 0);
		} else if (!strcasecmp(v->name, "echocancel")) {
				GLOB(echocancel) = ast_true(v->value);
		} else if (!strcasecmp(v->name, "silencesuppression")) {
			GLOB(silencesuppression) = ast_true(v->value);
		} else if (!strcasecmp(v->name, "rtptos")) {
			if (sscanf(v->value, "%d", &GLOB(rtptos)) == 1)
				GLOB(rtptos) &= 0xff;
		} else {
			ast_log(LOG_WARNING, "Unknown Key in [general]: %s\n", v->name);
		}
		v = v->next;
	}

	ast_codec_pref_string(&GLOB(global_codecs), pref_buf, sizeof(pref_buf) - 1);
	ast_verbose(VERBOSE_PREFIX_3 "GLOBAL: Preferred capability %s\n", pref_buf);

	if (!ntohs(GLOB(bindaddr.sin_port))) {
		GLOB(bindaddr.sin_port) = ntohs(DEFAULT_SCCP_PORT);
	}
	GLOB(bindaddr.sin_family) = AF_INET;

	v = ast_variable_browse(cfg, "devices");
	if (!v) {
		ast_log(LOG_WARNING, "Missing [devices] section, SCCP disabled\n");
		return 0;
	}
	d = build_device();
	/* read devices section */
	while (v) {
		if (!strcasecmp(v->name, "device")) {
			if ( (strlen(v->value) == 15) && (strncmp(v->value, "SEP",3) == 0) ) {
				sccp_copy_string(d->id, v->value, sizeof(d->id));
				ast_verbose(VERBOSE_PREFIX_3 "Added device '%s' (%s)\n", d->id, d->config_type);
				ast_mutex_lock(&GLOB(devices_lock));
				d->next = GLOB(devices);
				GLOB(devices) = d;
				ast_mutex_unlock(&GLOB(devices_lock));
			} else {
				ast_log(LOG_WARNING, "Wrong device param: %s => %s\n", v->name, v->value);
				free(d);
			}
			d = build_device();
			speeddial_index = 1;
			k_last = NULL;
		} else if (!strcasecmp(v->name, "type")) {
			sccp_copy_string(d->config_type, v->value, sizeof(d->config_type));
		} else if (!strcasecmp(v->name, "tzoffset")) {
			/* timezone offset */
			d->tz_offset = atoi(v->value);
		} else if (!strcasecmp(v->name, "autologin")) {
			sccp_copy_string(d->autologin, v->value, sizeof(d->autologin));
		} else if (!strcasecmp(v->name, "description")) {
			sccp_copy_string(d->description, v->value, sizeof(d->description));
		} else if (!strcasecmp(v->name, "allow")) {
			ast_parse_allow_disallow(&d->codecs, &d->capability, v->value, 1);
		} else if (!strcasecmp(v->name, "disallow")) {
			ast_parse_allow_disallow(&d->codecs, &d->capability, v->value, 0);
		} else if (!strcasecmp(v->name, "transfer")) {
			d->transfer = ast_true(v->value);
	#ifdef CS_SCCP_PARK
		} else if (!strcasecmp(v->name, "park")) {
			d->park = ast_true(v->value);
	#endif
		} else if (!strcasecmp(v->name, "speeddial")) {
			if (ast_strlen_zero(v->value)) {
				speeddial_index++;
				ast_verbose(VERBOSE_PREFIX_3 "Added empty speeddial\n");
			} else {
				sccp_copy_string(k_speed, v->value, sizeof(k_speed));
				splitter = k_speed;
				k_exten = strsep(&splitter, ",");
				k_name = strsep(&splitter, ",");
				k_hint = splitter;
				if (k_exten)
					ast_strip(k_exten);
				if (k_name)
					ast_strip(k_name);
				if (k_hint)
					ast_strip(k_hint);
				if (k_exten && !ast_strlen_zero(k_exten)) {
					if (!k_name)
						k_name = k_exten;
					k = malloc(sizeof(sccp_speed_t));
					if (!k)
						ast_log(LOG_WARNING, "Error allocating speedial %s => %s\n", v->name, v->value);
					else {
						memset(k, 0, sizeof(sccp_speed_t));
						sccp_copy_string(k->name, k_name, sizeof(k->name));
						sccp_copy_string(k->ext, k_exten, sizeof(k->ext));
						if (k_hint)
							sccp_copy_string(k->hint, k_hint, sizeof(k->hint));
						k->instance = speeddial_index++;
						if (!d->speed_dials)
							d->speed_dials = k;
						if (!k_last)
							k_last = k;
						else {
							k_last->next = k;
							k_last = k;
						}
						ast_verbose(VERBOSE_PREFIX_3 "Added speeddial %d: %s (%s)\n", k->instance, k->name, k->ext);
					}
				} else
					ast_log(LOG_WARNING, "Wrong speedial syntax: %s => %s\n", v->name, v->value);
			}
		} else {
			ast_log(LOG_WARNING, "Unknown Key in [devices]: %s\n", v->name);
		}
		v = v->next;
	}

	if (d) {
		sccp_dev_speeddials_delete(d);
		free(d);
		d = NULL;
	}

	v = ast_variable_browse(cfg, "lines");
	if (!v) {
		ast_log(LOG_WARNING, "Missing [lines] section, SCCP disabled\n");
		return 0;
	}

	l = build_line();
	while(v) {
		if (!strcasecmp(v->name, "line")) {
			if ( !ast_strlen_zero(v->value) ) {
				sccp_copy_string(l->name, ast_strip(v->value), sizeof(l->name));
				if (sccp_line_find_byname(v->value)) {
					ast_log(LOG_WARNING, "The line %s already exists\n", l->name);
					free(l);
				} else {
					ast_verbose(VERBOSE_PREFIX_3 "Added line '%s'\n", l->name);
					ast_mutex_lock(&GLOB(lines_lock));
					l->next = GLOB(lines);
					if (l->next)
						l->next->prev = l;
					GLOB(lines) = l;
					ast_mutex_unlock(&GLOB(lines_lock));
				}
			} else {
				ast_log(LOG_WARNING, "Wrong line param: %s => %s\n", v->name, v->value);
				free(l);
			}
			l = build_line();
		} else if (!strcasecmp(v->name, "id")) {
			sccp_copy_string(l->id, v->value, sizeof(l->id));
		} else if (!strcasecmp(v->name, "pin")) {
			sccp_copy_string(l->pin, v->value, sizeof(l->pin));
		} else if (!strcasecmp(v->name, "label")) {
			sccp_copy_string(l->label, v->value, sizeof(l->label));
		} else if (!strcasecmp(v->name, "description")) {
			sccp_copy_string(l->description, v->value, sizeof(l->description));
		} else if (!strcasecmp(v->name, "context")) {
			sccp_copy_string(l->context, v->value, sizeof(l->context));
		} else if (!strcasecmp(v->name, "cid_name")) {
			sccp_copy_string(l->cid_name, v->value, sizeof(l->cid_name));
		} else if (!strcasecmp(v->name, "cid_num")) {
			sccp_copy_string(l->cid_num, v->value, sizeof(l->cid_num));
		} else if (!strcasecmp(v->name, "callerid")) {
			ast_log(LOG_WARNING, "obsolete callerid param. Use cid_num and cid_name\n");
		} else if (!strcasecmp(v->name, "mailbox")) {
			sccp_copy_string(l->mailbox, v->value, sizeof(l->mailbox));
		} else if (!strcasecmp(v->name, "vmnum")) {
			sccp_copy_string(l->vmnum, v->value, sizeof(l->vmnum));
		} else if (!strcasecmp(v->name, "incominglimit")) {
			l->incominglimit = atoi(v->value);
			if (l->incominglimit < 1)
			l->incominglimit = 1;
			/* this is the max call phone limits. Just a guess. It's not important */
			if (l->incominglimit > 10)
			l->incominglimit = 10;
		} else if (!strcasecmp(v->name, "echocancel")) {
				l->echocancel = ast_true(v->value);
		} else if (!strcasecmp(v->name, "silencesuppression")) {
			l->silencesuppression = ast_true(v->value);
		} else if (!strcasecmp(v->name, "rtptos")) {
			if (sscanf(v->value, "%d", &l->rtptos) == 1)
				l->rtptos &= 0xff;
		} else if (!strcasecmp(v->name, "hint")) {
			sccp_copy_string(l->hint, v->value, sizeof(l->hint));
		}
		v = v->next;
	}

	if (l) {
		free(l);
	}

	/* ok the config parse is done */

	if ((GLOB(descriptor) > -1) && (ntohs(GLOB(bindaddr.sin_port)) != oldport)) {
		close(GLOB(descriptor));
		GLOB(descriptor) = -1;
	}

  if (GLOB(descriptor) < 0) {

	GLOB(descriptor) = socket(AF_INET, SOCK_STREAM, 0);

	on = 1;
	if (setsockopt(GLOB(descriptor), SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
		ast_log(LOG_WARNING, "Failed to set SCCP socket to SO_REUSEADDR mode: %s\n", strerror(errno));
	if (setsockopt(GLOB(descriptor), IPPROTO_IP, IP_TOS, &tos, sizeof(tos)) < 0)
		ast_log(LOG_WARNING, "Failed to set SCCP socket TOS to IPTOS_LOWDELAY: %s\n", strerror(errno));
	if (setsockopt(GLOB(descriptor), SOL_TCP, TCP_NODELAY, &on, sizeof(on)) < 0)
		ast_log(LOG_WARNING, "Failed to set SCCP socket to TCP_NODELAY: %s\n", strerror(errno));

	if (GLOB(descriptor) < 0) {

		ast_log(LOG_WARNING, "Unable to create SCCP socket: %s\n", strerror(errno));

	} else {
		if (bind(GLOB(descriptor), (struct sockaddr *)&GLOB(bindaddr), sizeof(GLOB(bindaddr))) < 0) {
		ast_log(LOG_WARNING, "Failed to bind to %s:%d: %s!\n",
		ast_inet_ntoa(iabuf, sizeof(iabuf), GLOB(bindaddr.sin_addr)), ntohs(GLOB(bindaddr.sin_port)),
		strerror(errno));
		close(GLOB(descriptor));
		GLOB(descriptor) = -1;
		return 0;
		}
		ast_verbose(VERBOSE_PREFIX_3 "SCCP channel driver up and running on %s:%d\n",
		ast_inet_ntoa(iabuf, sizeof(iabuf), GLOB(bindaddr.sin_addr)), ntohs(GLOB(bindaddr.sin_port)));

		if (listen(GLOB(descriptor), DEFAULT_SCCP_BACKLOG)) {
		ast_log(LOG_WARNING, "Failed to start listening to %s:%d: %s\n",
		ast_inet_ntoa(iabuf, sizeof(iabuf), GLOB(bindaddr.sin_addr)), ntohs(GLOB(bindaddr.sin_port)),
		strerror(errno));
		close(GLOB(descriptor));
		GLOB(descriptor) = -1;
		return 0;
		}

		sccp_log(0)(VERBOSE_PREFIX_3 "SCCP listening on %s:%d\n", ast_inet_ntoa(iabuf, sizeof(iabuf), GLOB(bindaddr.sin_addr)), ntohs(GLOB(bindaddr.sin_port)));
		ast_pthread_create(&socket_thread,NULL, sccp_socket_thread, NULL);
	}
  }

  ast_config_destroy(cfg);
  sccp_dev_dbclean();
  return 0;
}

static char *sccp_setcalledparty_descrip = "  SetCalledParty(\"Name\" <ext>) sets the name and number of the called party for use with chan_sccp\n";

static int sccp_setcalledparty_exec(struct ast_channel *chan, void *data) {
  char tmp[256] = "";
  char * num, * name;
  sccp_channel_t * c = CS_AST_CHANNEL_PVT(chan);

  if (strcasecmp(chan->type, "SCCP") != 0)
	return 0;

  if (!data || !c)
	return 0;

  sccp_copy_string(tmp, (char *)data, sizeof(tmp));
  ast_callerid_parse(tmp, &name, &num);

  sccp_channel_set_calledparty(c, name, num);

  return 0;
}

int reload(void) {
	return 0;
}

int load_module() {

  /* make globals */
  sccp_globals = malloc(sizeof(struct sccp_global_vars));
  memset(sccp_globals,0,sizeof(struct sccp_global_vars));
  ast_mutex_init(&GLOB(lock));
  ast_mutex_init(&GLOB(sessions_lock));
  ast_mutex_init(&GLOB(devices_lock));
  ast_mutex_init(&GLOB(intercoms_lock));
  ast_mutex_init(&GLOB(lines_lock));
  ast_mutex_init(&GLOB(channels_lock));
  ast_mutex_init(&GLOB(usecnt_lock));

  /* GLOB() is a macro for sccp_globals-> */

  GLOB(descriptor) = -1;
  GLOB(ourport) = 2000;
  GLOB(keepalive)  = SCCP_KEEPALIVE;
  sccp_copy_string(GLOB(date_format), "D/M/YA", sizeof(GLOB(date_format)));
  sccp_copy_string(GLOB(default_context), "default", sizeof(GLOB(default_context)));

/* Wait up to 16 seconds for first digit */
  GLOB(firstdigittimeout) = 16;
/* How long to wait for following digits */
  GLOB(digittimeout) = 8;
/* Yes, these are all that the phone supports (except it's own 'Wideband 256k') */
  GLOB(global_capability) = AST_FORMAT_ALAW|AST_FORMAT_ULAW|AST_FORMAT_G729A;

  GLOB(debug) = 1;
  GLOB(rtptos) = (184 & 0xff);
  GLOB(echocancel) = 1;
  GLOB(silencesuppression) = 0;

  if (!reload_config()) {
#ifdef CS_AST_HAS_TECH_PVT
	if (ast_channel_register(&sccp_tech)) {
#else
	if (ast_channel_register_ex("SCCP", "SCCP", GLOB(global_capability), sccp_request , NULL /* sccp_devicestate */)) {
#endif
		ast_log(LOG_ERROR, "Unable to register channel class SCCP\n");
		return -1;
	}
  }

  sccp_register_cli();
  ast_register_application("SetCalledParty", sccp_setcalledparty_exec, "Sets the name of the called party", sccp_setcalledparty_descrip);

  return 0;
}

int unload_module() {
	sccp_line_t * l;
	sccp_device_t * d;
	sccp_intercom_t * i;
	sccp_session_t * s;
	sccp_hint_t *h;
	char iabuf[INET_ADDRSTRLEN];

#ifdef CS_AST_HAS_TECH_PVT
	ast_channel_unregister(&sccp_tech);
#else
	ast_channel_unregister("SCCP");
#endif
	ast_unregister_application("SetCalledParty");
	sccp_unregister_cli();

	ast_mutex_lock(&GLOB(channels_lock));
	while (GLOB(channels))
		sccp_channel_delete(GLOB(channels));
	ast_mutex_unlock(&GLOB(channels_lock));

	ast_mutex_lock(&GLOB(lines_lock));
	while (GLOB(lines)) {
		l = GLOB(lines);
		GLOB(lines) = l->next;
		sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Removing line %s\n", l->name);
		while (l->hints) {
			h = l->hints;
			l->hints = l->hints->next;
			free(h);
		}
		free(l);
	}
	ast_mutex_unlock(&GLOB(lines_lock));

	ast_mutex_lock(&GLOB(devices_lock));
	while (GLOB(devices)) {
		d = GLOB(devices);
		GLOB(devices) = d->next;
		sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Removing device %s\n", d->id);
		sccp_dev_speeddials_delete(d);
		if (d->btntemplate)
			free(d->btntemplate);
		while (d->hints) {
			h = d->hints;
			d->hints = d->hints->next;
			free(h);
		}
		free(d);
	}
	ast_mutex_unlock(&GLOB(devices_lock));

	ast_mutex_lock(&GLOB(intercoms_lock));
	while (GLOB(intercoms)) {
		i = GLOB(intercoms);
		GLOB(intercoms) = i->next;
		sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Removing intercom %s\n", i->id);
		free(i);
	}
	ast_mutex_unlock(&GLOB(intercoms_lock));

	ast_mutex_lock(&GLOB(sessions_lock));

	while (GLOB(sessions)) {
		s = GLOB(sessions);
		GLOB(sessions) = s->next;
		sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Removing session %s\n", ast_inet_ntoa(iabuf, sizeof(iabuf), s->sin.sin_addr));
		if (s->fd > -1)
			close(s->fd);
		free(s);
	}
	ast_mutex_unlock(&GLOB(sessions_lock));
	close(GLOB(descriptor));
	GLOB(descriptor) = -1;

	if (!ast_mutex_lock(&GLOB(socket_lock))) {
		if (socket_thread && (socket_thread != AST_PTHREADT_STOP)) {
			pthread_cancel(socket_thread);
			pthread_kill(socket_thread, SIGURG);
			pthread_join(socket_thread, NULL);
		}
		socket_thread = AST_PTHREADT_STOP;
		ast_mutex_unlock(&GLOB(socket_lock));
	} else {
		ast_log(LOG_WARNING, "SCCP: Unable to lock the socket\n");
		return -1;
	}

	free(sccp_globals);

	return 0;
}

int usecount() {
	int res;
	ast_mutex_lock(&GLOB(usecnt_lock));
	res = GLOB(usecnt);
	ast_mutex_unlock(&GLOB(usecnt_lock));
	return res;
}

char *key() {
	return ASTERISK_GPL_KEY;
}

char *description() {
	return ("Skinny Client Control Protocol (SCCP). Release: " SCCP_VERSION);
}
