#ifndef __CHAN_SCCP_H
#define __CHAN_SCCP_H

#include "config.h"
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <asterisk/frame.h>
#include <asterisk/module.h>
#include <asterisk/channel.h>
#include <asterisk/lock.h>
#include <asterisk/logger.h>
#include <asterisk/sched.h>
#include <asterisk/io.h>
#include <asterisk/rtp.h>
#include <asterisk/config.h>
#include <asterisk/options.h>
#include <asterisk/utils.h>
#ifndef CS_AST_HAS_TECH_PVT
#include <asterisk/channel_pvt.h>
#endif

#ifdef linux
#include <endian.h>
#endif /* linux */

#ifdef __FreeBSD__
#include <sys/endian.h>
#endif /* FreeBSD */

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define letohl(x) (x)
#define letohs(x) (x)
#define htolel(x) (x)
#define htoles(x) (x)
#else
#include <bits/byteswap.h>
#define letohl(x) __bswap_32(x)
#define letohs(x) __bswap_16(x)
#define htolel(x) __bswap_32(x)
#define htoles(x) __bswap_16(x)
#endif

#define SCCP_VERSION "20050826"

#define sccp_log(x) if (sccp_globals->debug >= x) ast_verbose
#define GLOB(x) sccp_globals->x

#ifdef CS_AST_HAS_TECH_PVT
#define CS_AST_CHANNEL_PVT(x) x->tech_pvt
#else
#define CS_AST_CHANNEL_PVT(x) x->pvt->pvt
#endif

#ifdef CS_AST_HAS_BRIDGED_CHANNEL
#define CS_AST_BRIDGED_CHANNEL(x) ast_bridged_channel(x)
#else
#define CS_AST_BRIDGED_CHANNEL(x) x->bridge
#endif

typedef struct sccp_channel		sccp_channel_t;
typedef struct sccp_session		sccp_session_t;
typedef struct sccp_line		sccp_line_t;
typedef struct sccp_speed		sccp_speed_t;
typedef struct sccp_device		sccp_device_t;
typedef struct sccp_intercom	sccp_intercom_t;
typedef struct sccp_pvt			sccp_pvt_t;
typedef struct sccp_hint		sccp_hint_t;

typedef void sk_func (sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c);

#include "sccp_protocol.h"

/* An asterisk channel can now point to a channel (single device)
 * or an intercom.	So we need somethign to pass around :) */
struct sccp_pvt {
  sccp_channel_t  * chan;
  sccp_intercom_t * icom;
};

struct sccp_hint {
	int hintid;
	sccp_device_t  *device;
	uint8_t	instance;
	sccp_hint_t *next;
};

/* A line is a the equiv of a 'phone line' going to the phone. */
struct sccp_line {

  /* lockmeupandtiemedown */
  ast_mutex_t lock;

  /* This line's ID, used for logging into (for mobility) */
  char id[4];

  /* PIN number for mobility/roaming. */
  char pin[8];

  /* The lines position/instanceId on the current device*/
  uint8_t instance;

  /* the name of the line, so use in asterisk (i.e SCCP/<name>) */
  char name[80];

  /* A description for the line, displayed on in header (on7960/40)
   * or on main  screen on 7910 */
  char description[StationMaxNameSize];

  /* A name for the line, displayed next to the button (7960/40). */
  char label[StationMaxNameSize];

  /* mainbox numbers (seperated by commas) to check for messages */
  char mailbox[AST_MAX_EXTENSION];

  /* Voicemail number to dial */
  char vmnum[AST_MAX_EXTENSION];

  /* The context we use for outgoing calls. */
  char context[AST_MAX_CONTEXT];

  /* CallerId to use on outgoing calls*/
  char cid_name[AST_MAX_EXTENSION];
  char cid_num[AST_MAX_EXTENSION];

  /* max incoming calls limit */
  uint8_t incominglimit;

  /* rtp stream tos */
  uint32_t	rtptos;

  /* The currently active channel. */
/*  sccp_channel_t * activeChannel; */

  /* Linked list of current channels for this line */
  sccp_channel_t * channels;

  /* Number of currently active channels */
  uint8_t channelCount;

  /* Prev and Next line in the global list */
  sccp_line_t * next, * prev;

  /* Prev and Next line on the current device. */
  sccp_line_t * next_on_device, * prev_on_device;

  /* The device this line is currently registered to. */
  sccp_device_t * device;

  /* hint system */
  char hint[AST_MAX_EXTENSION];
  /* list of hint pointers. Internal lines to notify the state */
  sccp_hint_t * hints;

  unsigned int		   hasMessages:1;

  /* echocancel phone support */
  unsigned int		   echocancel:1;
  unsigned int		   silencesuppression:1;
  unsigned int		   spareBit3:1;
  unsigned int		   spareBit4:1;
  unsigned int		   spareBit5:1;
  unsigned int		   spareBit6:1;

};

/* This defines a speed dial button */
struct sccp_speed {
  /* The instance on the current device */
  int instance;

  /* The name of the speed dial button */
  char name[StationMaxNameSize];

  /* The number to dial when it's hit */
  char ext[AST_MAX_EXTENSION];
  char hint[AST_MAX_EXTENSION];

  /* Pointer to next speed dial */
  sccp_speed_t * next;

};


struct sccp_device {
  ast_mutex_t lock;

  /* SEP<macAddress> of the device. */
  char id[StationMaxDeviceNameSize];

  /* internal description. Skinny protocol does not use it */
  char description[40];

  /* lines to auto login this device to */
  char autologin[SCCP_MAX_AUTOLOGIN];

  /* model of this phone used for setting up features/softkeys/buttons etc. */
  char config_type[10];

  /* model of this phone sent by the station, devicetype*/
  uint32_t skinny_type;

  /* timezone offset */
  int tz_offset;

  /* version to send to the phone */
  char imgversion[24];

  /* If the device has been rully registered yet */
  uint8_t registrationState;

  /* time() the last call ended. */
  time_t lastCallEndTime;

  /* asterisk codec device preference */
  struct ast_codec_pref codecs;

  /* SCCP_DEVICE_ONHOOK or SCCP_DEVICE_OFFHOOK */
  uint8_t state;
  
  /* ringer mode. Need it for the ringback */
  uint8_t ringermode;

  /* last dialed number */
  char lastNumber[AST_MAX_EXTENSION];

  /* asterisk codec capability */
  int capability;

  btnlist * btntemplate;

  unsigned int		   hasMessages:1;
  unsigned int		   dnd:1;
  unsigned int		   transfer:1;
  unsigned int		   park:1;
  unsigned int		   spareBit4:1;
  unsigned int		   spareBit5:1;
  unsigned int		   spareBit6:1;
  unsigned int		   spareBit7:1;

  sccp_channel_t * active_channel;
  /* the channel under transfer */
  sccp_channel_t * transfer_channel;
  sccp_speed_t * speed_dials;
  sccp_line_t * lines;
  sccp_line_t * currentLine;
  sccp_session_t * session;
  sccp_device_t * next;
  /* list of hint pointers. Internal lines to notify the state */
  sccp_hint_t * hints;
};

struct sccp_intercom {
  ast_mutex_t		lock;
  char				id[24];
  char				description[24];
  sccp_device_t  ** devices;
  sccp_intercom_t * next;
};

struct sccp_session {
  ast_mutex_t			lock;
  void					*buffer;
  size_t				buffer_size;
  struct sockaddr_in	sin;
  /* ourip is for rtp use */
  struct in_addr		ourip;
  time_t				lastKeepAlive;
  int					fd;
  int					rtpPort;
  sccp_device_t 		*device;
  sccp_session_t		*prev, *next;
};

struct sccp_channel {
  ast_mutex_t			lock;
  char					calledPartyName[StationMaxNameSize];
  char					calledPartyNumber[StationMaxDirnumSize];
  char					callingPartyName[StationMaxNameSize];
  char					callingPartyNumber[StationMaxDirnumSize];
  uint32_t				callid;
  /* internal channel state SCCP_CHANNELSTATE_* */
  uint8_t				state;
  /* skinny state */
  uint8_t				callstate;
  /* SKINNY_CALLTYPE_* */
  uint8_t				calltype;
  /* timeout on dialing state */
  time_t				digittimeout;
  /* SCCPRingerMode application */
  uint8_t				ringermode;
  sccp_device_t 	 *	device;
  struct ast_channel *	owner;
  sccp_line_t		 *	line;
  struct ast_rtp	 *	rtp;
  struct sockaddr_in	rtp_addr;
  sccp_channel_t	 *	prev, * next,
					 *	prev_on_line, * next_on_line;
};

struct sccp_global_vars {
	ast_mutex_t				lock;

	sccp_session_t			*sessions;
	ast_mutex_t				sessions_lock;
	sccp_device_t			*devices;
	ast_mutex_t				devices_lock;
	sccp_intercom_t 		*intercoms;
	ast_mutex_t				intercoms_lock;
	sccp_line_t	  			*lines;
	ast_mutex_t				lines_lock;
	sccp_channel_t 			*channels;
	ast_mutex_t				channels_lock;
	ast_mutex_t				socket_lock;
	int 					descriptor;
	/* Keep track of when we're in use. */
	int						usecnt;
	ast_mutex_t				usecnt_lock;
	struct sockaddr_in		bindaddr;
	int						ourport;

	char					default_context[AST_MAX_CONTEXT];
	int						global_capability;
	struct					ast_codec_pref global_codecs;
	int						keepalive;
	int						debug;
	char 					date_format[7];
/* Wait up to 16 seconds for first digit */
	uint8_t					firstdigittimeout;
/* How long to wait for following digits */
	uint8_t					digittimeout;
	uint32_t				rtptos;
	/* echocancel phone support */
	unsigned int		   echocancel:1;
	unsigned int		   silencesuppression:1;
	unsigned int		   spareBit2:1;
	unsigned int		   spareBit3:1;
	unsigned int		   spareBit4:1;
	unsigned int		   spareBit5:1;
	unsigned int		   spareBit6:1;
};

struct sccp_global_vars *sccp_globals;

int sccp_handle_message(sccp_moo_t * r, sccp_session_t * s);

#ifdef CS_AST_HAS_TECH_PVT
struct ast_channel *sccp_request(const char *type, int format, void *data, int *cause);
#else
struct ast_channel *sccp_request(char *type, int format, void *data);
#endif

int sccp_devicestate(void *data);
sccp_hint_t * sccp_hint_make(sccp_device_t *d, uint8_t instance);
int sccp_hint_state(char *context, char* exten, int state, void *data);
void sccp_hint_notify(sccp_channel_t *c);

#endif /* __CHAN_SCCP_H */

