/*
 * (SCCP*)
 *
 * An implementation of Skinny Client Control Protocol (SCCP)
 *
 * Sergio Chersovani (mlists@c-net.it)
 *
 * Reworked, but based on chan_sccp code.
 * The original chan_sccp driver that was made by Zozo which itself was derived from the chan_skinny driver.
 * Modified by Jan Czmok and Julien Goodwin
 *
 * This program is free software and may be modified and
 * distributed under the terms of the GNU Public License.
 */

#include "config.h"
#include "chan_sccp.h"
#include "sccp_actions.h"
#include "sccp_utils.h"
#include "sccp_pbx.h"
#include "sccp_channel.h"
#include "sccp_indicate.h"
#include "sccp_device.h"
#include "sccp_line.h"
#include "sccp_socket.h"
#include <asterisk/utils.h>
#include <asterisk/pbx.h>

void sccp_handle_alarm(sccp_session_t * s, sccp_moo_t * r) {
	sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Alarm Message: Severity: %s (%d), %s [%d/%d]\n", skinny_alarm2str(letohl(r->msg.AlarmMessage.lel_alarmSeverity)), letohl(r->msg.AlarmMessage.lel_alarmSeverity), r->msg.AlarmMessage.text, letohl(r->msg.AlarmMessage.lel_parm1), letohl(r->msg.AlarmMessage.lel_parm2));
}

void sccp_handle_register(sccp_session_t * s, sccp_moo_t * r) {
	sccp_device_t * d;
	btnlist btn[StationMaxButtonTemplateSize];
	char *mb, *cur, tmp[256];
	sccp_line_t *l, *lines_last = NULL;
	sccp_moo_t * r1;
	char iabuf[INET_ADDRSTRLEN];
	uint8_t i = 0, line_count = 0;
	memset(&btn, 0 , sizeof(btn));

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: is registering, Instance: %d, Type: %s (%d), Version: %d\n",
		r->msg.RegisterMessage.sId.deviceName,
		letohl(r->msg.RegisterMessage.sId.lel_instance),
		skinny_devicetype2str(letohl(r->msg.RegisterMessage.lel_deviceType)),
		letohl(r->msg.RegisterMessage.lel_deviceType),
		r->msg.RegisterMessage.protocolVer);

	d = sccp_device_find_byid(r->msg.RegisterMessage.sId.deviceName);
	if (!d) {
		REQ(r1, RegisterRejectMessage);
		ast_log(LOG_ERROR, "%s: Rejecting device: not found\n", r1->msg.RegisterMessage.sId.deviceName);
		sccp_copy_string(r1->msg.RegisterRejectMessage.text, "Unknown Device", sizeof(r1->msg.RegisterRejectMessage.text));
		sccp_session_send(s, r1);
		return;
	}
	ast_mutex_lock(&d->lock);
	if (d->session) {
		ast_log(LOG_WARNING, "%s: Device is doing a re-registration!\n", d->id);
	}
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Allocating device to session %s\n", d->id, ast_inet_ntoa(iabuf, sizeof(iabuf), s->sin.sin_addr));
	s->device = d;
	d->skinny_type = letohl(r->msg.RegisterMessage.lel_deviceType);
	d->session = s;
	s->lastKeepAlive = time(0);
	ast_mutex_unlock(&d->lock);

	/* pre-attach lines. We will wait for button template req if the phone does support it */
	sccp_dev_build_buttontemplate(d, btn);

	line_count = 0;
	/* count the available lines on the phone */
	for (i = 0; i < StationMaxButtonTemplateSize; i++) {
		if ( (btn[i].type == SKINNY_BUTTONTYPE_LINE) || (btn[i].type == SKINNY_BUTTONTYPE_MULTI) )
			line_count++;
		else if (btn[i].type == SKINNY_BUTTONTYPE_UNUSED)
			break;
	}
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Phone available lines %d\n", d->id, line_count);
	i = 0;
	sccp_copy_string(tmp, d->autologin, sizeof(tmp));
	mb = tmp;
	while((cur = strsep(&mb, ","))) {
		ast_strip(cur);
		if (ast_strlen_zero(cur)) {
			/* this is useful to leave and empty button */
			i++;
			continue;
		}
		
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Auto logging into %s\n", d->id, cur);
		l = sccp_line_find_byname(cur);
		if (!l) {
			ast_log(LOG_ERROR, "%s: Failed to autolog into %s: Couldn't find line %s\n", d->id, cur, cur);
			i++;
			continue;
		}
		
		if (l->device) {
			sccp_log(10)(VERBOSE_PREFIX_3 "%s: Line %s aready attached to device %s\n", d->id, l->name, l->device->id);
			i++;
			continue;
		}
		
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Attaching line %s to this device\n", d->id, l->name);
		if (i == line_count) {
			ast_log(LOG_WARNING, "%s: Failed to autolog into %s: Max available lines phone limit reached %s\n", d->id, cur, cur);
			continue;
		}

		i++;
		ast_mutex_lock(&l->lock);
		ast_mutex_lock(&d->lock);
		if (i == 1)
		d->currentLine = l;
		l->device = d;
		l->instance = i;
		/* I want it in the right order */
		if (!d->lines)
			d->lines = l;
		if (!lines_last)
			lines_last = l;
		else {
			l->prev_on_device = lines_last;
			lines_last->next_on_device = l;
			lines_last = l;
		}
/*					if (!ast_strlen_zero(l->hint)) {
				l->hintid = ast_extension_state_add(l->context, l->hint, sccp_hint_state, l);
				if (l->hintid > -1) {
					sccp_log(10)(VERBOSE_PREFIX_3 "%s: Added exstension monitor on %s\n", d->id, l->hint);
				} else {
					sccp_log(10)(VERBOSE_PREFIX_3 "%s: Error adding exstension monitor on %s\n", d->id, l->hint);
				}
			}
*/
		ast_mutex_unlock(&l->lock);
		ast_mutex_unlock(&d->lock);
	}

	REQ(r1, RegisterAckMessage);
	r1->msg.RegisterAckMessage.lel_protocolVer = htolel(0x3);
	r1->msg.RegisterAckMessage.lel_keepAliveInterval = htolel(GLOB(keepalive));
	r1->msg.RegisterAckMessage.lel_secondaryKeepAliveInterval = htolel(GLOB(keepalive));
	memcpy(r1->msg.RegisterAckMessage.dateTemplate, GLOB(date_format), sizeof(r1->msg.RegisterAckMessage.dateTemplate));
	sccp_session_send(s, r1);
	sccp_dev_set_registered(d, SKINNY_DEVICE_RS_PROGRESS);
	sccp_dev_set_registered(d, SKINNY_DEVICE_RS_OK);

	sccp_dev_check_mwi(d);
	/* check for dnd and cfwd status */
	sccp_dev_dbget(d);
	if (d->dnd)
		sccp_dev_displayprompt(d, NULL, ">>> " SKINNY_DISP_DND " <<<", 0);

	sccp_dev_sendmsg(d, CapabilitiesReqMessage);
}


void sccp_handle_unregister(sccp_session_t * s, sccp_moo_t * r) {
	sccp_moo_t * r1;
	sccp_device_t * d = s->device;
	if (!s || (s->fd < 0) )
		return;

	/* we don't need to look for active channels. the phone does send unregister only when there are no channels */
	REQ(r1, UnregisterAckMessage);
  	r1->msg.UnregisterAckMessage.lel_status = SKINNY_UNREGISTERSTATUS_OK;
	sccp_session_send(s, r1);
	sccp_log(1)(VERBOSE_PREFIX_3 "%s: unregister request sent\n", (d) ? d->id : "SCCP");
	if (d)
		sccp_dev_set_registered(d, SKINNY_DEVICE_RS_NONE);
	sccp_session_close(s);
}


static btnlist *sccp_make_button_template(sccp_device_t * d) {
	int i;
	sccp_speed_t *k = NULL, *k1 = NULL;
	sccp_line_t *l = NULL, * l1 = NULL;
	sccp_hint_t *h = NULL;
	int lineInstance = 1, btn_count = 1;
	char hint[256] = "", hint_dialplan[256] = "";
	char *splitter, *hint_exten, *hint_context;
	btnlist *btn; 
	
	btn = malloc(sizeof(btnlist)*StationMaxButtonTemplateSize);
	if (!btn)
		return NULL;
		
	memset(btn, 0 , sizeof(btnlist)*StationMaxButtonTemplateSize);
	sccp_dev_build_buttontemplate(d, btn);
	memset(&hint_dialplan, 0, sizeof(hint_dialplan));

	ast_mutex_lock(&GLOB(devices_lock));
	ast_mutex_lock(&GLOB(lines_lock));
	ast_mutex_lock(&d->lock);
	
	/* line button template configuration */
	l = d->lines;
	btn_count = 1;
	/* Skinny line request message does not request the instance number, so we need to change it here */
	lineInstance = 1;
	while (l) {
		l->device = NULL;
		btn_count = 1;
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Looking for a line button place %s (%d)\n", d->id, l->name, l->instance);
		for (i = 0 ; i < StationMaxButtonTemplateSize ; i++) {
			if ( (btn[i].type == SKINNY_BUTTONTYPE_LINE) 
				|| ( btn[i].type == SKINNY_BUTTONTYPE_MULTI)
				|| ( btn[i].type == SCCP_BUTTONTYPE_LINE)) {
				if (btn_count == l->instance) {
					l->device = d;
					l->instance = i + 1;
					btn[i].instance = l->instance;
					btn[i].type = SCCP_BUTTONTYPE_LINE;
					btn[i].ptr = l;
					sccp_log(10)(VERBOSE_PREFIX_3 "%s: Configured Phone Button [%.2d] = LINE (%s) instance (%d)\n", d->id, i+1, l->name, l->instance);
					break;
				}
				btn_count++;
			}
		}
		l1 = l->next_on_device;
		if (!l->device) {
			/* unused line */
			sccp_log(10)(VERBOSE_PREFIX_3 "%s: removing unused line %s\n", d->id, l->name);
			sccp_line_delete_nolock(l);
		}
		l = l1;
	}

	/* Speedials with hint configuration */
	
	k1 = NULL;
	k = d->speed_dials;
	btn_count = 1;
	while (k) {
		btn_count = 1;
		h = NULL;
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Looking for a speeddial button place %s (%d)\n", d->id, k->name, k->instance);
/*		for (i = 0 ; i < StationMaxButtonTemplateSize ; i++) {
		if ( btn[i].type != 0)
			sccp_log(10)(VERBOSE_PREFIX_3 "%s: Button Template [%.2d] = %s (%d), instance %d\n", d->id, i+1, skinny_buttontype2str(btn[i].type), btn[i].type,btn[i].instance);
		} */
		for (i = 0 ; i < StationMaxButtonTemplateSize ; i++) {
			if ( (btn[i].type == SKINNY_BUTTONTYPE_SPEEDDIAL) || (btn[i].type == SKINNY_BUTTONTYPE_MULTI) || (btn[i].type == SCCP_BUTTONTYPE_HINT) || (btn[i].type == SCCP_BUTTONTYPE_SPEEDDIAL)) {
				if (btn_count == k->instance) {
					k->instance = i + 1;
					btn[i].instance = k->instance;
					btn[i].ptr = k;
					
					if (!ast_strlen_zero(k->hint)) {
						/* look for hints */
						sccp_log(10)(VERBOSE_PREFIX_3 "%s: Parsing speedial hint %s for speedial: %s,%s\n", d->id, k->hint, k->ext, k->name);
						sccp_copy_string(hint, k->hint, sizeof(k->hint));
						splitter = hint;
						hint_exten = strsep(&splitter, "@");
						if (hint_exten)
							ast_strip(hint_exten);
						hint_context = splitter;
						if (hint_context)
							ast_strip(hint_context);
						else
							hint_context = GLOB(default_context);
#ifdef CS_AST_HAS_NEW_HINT
						ast_get_hint(hint_dialplan, sizeof(hint_dialplan) - 1, NULL, 0, NULL, hint_context, hint_exten);
#else
						ast_get_hint(hint_dialplan, sizeof(hint_dialplan) - 1, NULL, hint_context, hint_exten);
#endif
						if (ast_strlen_zero(hint_dialplan)) {
							sccp_log(10)(VERBOSE_PREFIX_3 "%s: No hint configuration in the dialplan %s for exten: %s and context: %s\n", d->id, hint_dialplan, hint_exten, hint_context);
						} else {
							h = sccp_hint_make(d, k->instance);
							if (!h) {
								sccp_log(10)(VERBOSE_PREFIX_3 "%s: Error adding speeddial hint %s for extension %s and context: %s\n", d->id, hint_dialplan, hint_exten, hint_context);
							} else {
								h->next = d->hints;
								d->hints = h;
								if ( strchr(hint_dialplan,'&') || strncasecmp(hint_dialplan,"SCCP",4) ) {
									h->hintid = ast_extension_state_add(hint_context, hint_exten, sccp_hint_state, h);
									sccp_log(10)(VERBOSE_PREFIX_3 "%s: Configuring asterisk (no sccp features) hint %s for exten: %s and context: %s\n", d->id, hint_dialplan, hint_exten, hint_context);
								} else {
									l = NULL;
									splitter = hint_dialplan;
									strsep(&splitter, "/");
									sccp_copy_string(hint_dialplan, splitter, sizeof(hint_dialplan));
									if (hint_dialplan)
										ast_strip(hint_dialplan);
									if (hint_dialplan)
										l = sccp_line_find_byname(hint_dialplan);
									if (!l) {
										sccp_log(10)(VERBOSE_PREFIX_3 "%s: Error adding hint for line: %s\n", d->id, hint_dialplan);
									} else {
										h = sccp_hint_make(d, k->instance);
										if (!h) {
											sccp_log(10)(VERBOSE_PREFIX_3 "%s: Error creating hint for line: %s\n", d->id, hint_dialplan);
										} else {
											h->next = l->hints;
											l->hints = h;
											sccp_log(10)(VERBOSE_PREFIX_3 "%s: Added hint on line %s (%p) for extension %s and context: %s\n", d->id, l->name, l,hint_exten, hint_context);
										}
									}
								}
							}
						}
					}

					if (h)
						btn[i].type = SCCP_BUTTONTYPE_HINT;
					else
						btn[i].type = SCCP_BUTTONTYPE_SPEEDDIAL;
					sccp_log(10)(VERBOSE_PREFIX_3 "%s: Configured Phone Button [%.2d] = %s (%s) instance (%d)\n", d->id, i+1, (h) ? "SPEEDIAL WITH HINT" : "SPEEDIAL",k->name, k->instance);
					break;
				}
				btn_count++;
			}
			if ( i == (StationMaxButtonTemplateSize - 1) && (k->instance != i) ) {
				sccp_log(10)(VERBOSE_PREFIX_3 "%s: removing unused speedial %s\n", d->id, k->name);
				if (k1) {
					k1->next = k->next;
					free(k);
					k = k1;
				} else {
					d->speed_dials = NULL;
					free(k);
					k = NULL;
				}
			}
		}
		k1 = k;
		if (k)
			k = k->next;
	}

	/* cleanups and count */
	btn_count = 0;
	/* cleanup the multi unused buttons 7914 fix*/
	for (i = StationMaxButtonTemplateSize - 1; i >= 0 ; i--) {
		if (btn[i].instance == 0) {
			btn[i].type = SKINNY_BUTTONTYPE_UNDEFINED;
		} else {
			break;
		}
	}
	ast_mutex_unlock(&d->lock);
	ast_mutex_unlock(&GLOB(lines_lock));
	ast_mutex_unlock(&GLOB(devices_lock));
	return btn;
}
// Handle the configuration of the device here.

void sccp_handle_button_template_req(sccp_session_t * s, sccp_moo_t * r) {
	btnlist *btn;
	sccp_device_t * d = s->device;
	int i;
	sccp_moo_t * r1;
	
	if (!d)
		return;

	if (d->registrationState != SKINNY_DEVICE_RS_OK) {
		ast_log(LOG_WARNING, "%s: Received a button template request from unregistered device\n", d->id);
		sccp_session_close(s);
		return;
	}
	


	if (!d->btntemplate) {
		d->btntemplate = sccp_make_button_template(d);
		if (!d->btntemplate) {
			ast_log(LOG_ERROR, "%s: No memory allocated for button template\n", d->id);
			sccp_session_close(s);
			return;
		}
	}
	
	btn = d->btntemplate;
	REQ(r1, ButtonTemplateMessage);
	for (i = 0; i < StationMaxButtonTemplateSize ; i++) {
		if (btn[i].type == SCCP_BUTTONTYPE_HINT || btn[i].type == SCCP_BUTTONTYPE_LINE) {
			r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition = SKINNY_BUTTONTYPE_LINE;
			r1->msg.ButtonTemplateMessage.definition[i].instanceNumber = btn[i].instance;
		} else if (btn[i].type == SCCP_BUTTONTYPE_SPEEDDIAL) {
			r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition = SKINNY_BUTTONTYPE_SPEEDDIAL;
			r1->msg.ButtonTemplateMessage.definition[i].instanceNumber = btn[i].instance;
		} else if (btn[i].type == SKINNY_BUTTONTYPE_MULTI) {
			r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition = SKINNY_BUTTONTYPE_DISPLAY;
		} else if (btn[i].type != SKINNY_BUTTONTYPE_UNUSED) {
			r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition = btn[i].type;
		} else
			r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition = SKINNY_BUTTONTYPE_UNDEFINED;
			
		if (r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition != SKINNY_BUTTONTYPE_UNDEFINED) {
			r1->msg.ButtonTemplateMessage.lel_buttonCount++;
			sccp_log(10)(VERBOSE_PREFIX_3 "%s: Button Template [%.2d] = %s (%d), instance %d\n", d->id, i+1, skinny_buttontype2str(r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition), r1->msg.ButtonTemplateMessage.definition[i].buttonDefinition,r1->msg.ButtonTemplateMessage.definition[i].instanceNumber);
		}
	}
	
	r1->msg.ButtonTemplateMessage.lel_buttonOffset = htolel(0);
	r1->msg.ButtonTemplateMessage.lel_buttonCount = htolel(r1->msg.ButtonTemplateMessage.lel_buttonCount);
	r1->msg.ButtonTemplateMessage.lel_totalButtonCount = htolel(r1->msg.ButtonTemplateMessage.lel_buttonCount);
	sccp_dev_send(d, r1);
}

void sccp_handle_line_number(sccp_session_t * s, sccp_moo_t * r) {
	uint8_t lineNumber = letohs(r->msg.LineStatReqMessage.lel_lineNumber);
	sccp_line_t * l = NULL;
	sccp_moo_t * r1;
	sccp_device_t * d;
	sccp_speed_t * k = NULL;
	uint8_t i = 0, count = 0;
	if (!s)
		return;
	d = s->device;
	if (!d)
		return;

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Configuring line number %d\n", d->id, lineNumber);
	for (i = 0; i < StationMaxButtonTemplateSize ; i++) {
		if ( (d->btntemplate[i].type == SCCP_BUTTONTYPE_LINE) || (d->btntemplate[i].type == SCCP_BUTTONTYPE_HINT) ) {
			count++;
			if (count == lineNumber) {
				if (d->btntemplate[i].type == SCCP_BUTTONTYPE_LINE)
					l = d->btntemplate[i].ptr;
				else
					k = d->btntemplate[i].ptr;
				break;
			}
		}
	}
	REQ(r1, LineStatMessage);
	if (!l && !k) {
		ast_log(LOG_ERROR, "%s: requested a line configuration for unknown line %d\n", s->device->id, lineNumber);
		r1->msg.LineStatMessage.lel_lineNumber = htolel(lineNumber);
		sccp_dev_send(s->device, r1);
		return;
	}
	r1->msg.LineStatMessage.lel_lineNumber = htolel((l) ? l->instance : k->instance);
	
	sccp_copy_string(r1->msg.LineStatMessage.lineDirNumber, ((l) ? l->name : k->name), sizeof(r1->msg.LineStatMessage.lineDirNumber));
	sccp_copy_string(r1->msg.LineStatMessage.lineFullyQualifiedDisplayName, ((l) ? l->description : k->name), sizeof(r1->msg.LineStatMessage.lineFullyQualifiedDisplayName));
	sccp_copy_string(r1->msg.LineStatMessage.lineDisplayName, ((l) ? l->label : k->name), sizeof(r1->msg.LineStatMessage.lineDisplayName));
	sccp_dev_send(d, r1);
//	sccp_dev_set_keyset(d, lineNumber, 0, KEYMODE_ONHOOK);
}

void sccp_handle_speed_dial_stat_req(sccp_session_t * s, sccp_moo_t * r) {
	sccp_speed_t * k = s->device->speed_dials;
	sccp_moo_t * r1;
	int wanted = letohl(r->msg.SpeedDialStatReqMessage.lel_speedDialNumber);

	sccp_log(3)(VERBOSE_PREFIX_3 "%s: Speed Dial Request for Button %d\n", s->device->id, wanted);

	REQ(r1, SpeedDialStatMessage);
	r1->msg.SpeedDialStatMessage.lel_speedDialNumber = htolel(wanted);
	
	k = sccp_dev_speed_find_byindex(s->device, wanted);
	if (k) {
		sccp_copy_string(r1->msg.SpeedDialStatMessage.speedDialDirNumber, k->ext, sizeof(r1->msg.SpeedDialStatMessage.speedDialDirNumber));
		sccp_copy_string(r1->msg.SpeedDialStatMessage.speedDialDisplayName, k->name, sizeof(r1->msg.SpeedDialStatMessage.speedDialDisplayName));
	} else {
		sccp_log(3)(VERBOSE_PREFIX_3 "%s: speeddial %d not assigned\n", s->device->id, wanted);
	}

	sccp_dev_send(s->device, r1);
}

void sccp_handle_stimulus(sccp_session_t * s, sccp_moo_t * r) {
	sccp_device_t * d = s->device;
	sccp_line_t * l;
	sccp_speed_t * k;
	sccp_channel_t * c;
	uint8_t stimulus;
	uint8_t instance;

	if (!d || !r)
		return;

	k = d->speed_dials;
	stimulus = letohl(r->msg.StimulusMessage.lel_stimulus);
	instance = letohl(r->msg.StimulusMessage.lel_stimulusInstance);

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Got stimulus=%s (%d) for instance=%d\n", d->id, skinny_stimulus2str(stimulus), stimulus, instance);

	switch (stimulus) {

		case SKINNY_BUTTONTYPE_LASTNUMBERREDIAL: // We got a Redial Request
			if (!d->lastNumber)
				return;
			sccp_channel_newcall(d->currentLine, d->lastNumber);
			break;

		case SKINNY_BUTTONTYPE_LINE: // We got a Line Request
			l = sccp_line_find_byid(d, instance);
			if (!l) {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: No line for instance %d. Looking for a speedial with hint\n", d->id, instance);
				r->msg.StimulusMessage.lel_stimulus = htolel(SKINNY_BUTTONTYPE_SPEEDDIAL);
				sccp_handle_stimulus(s,r);
				return;
			}
			sccp_log(1)(VERBOSE_PREFIX_3 "%s: Line Key press on line %s\n", d->id, (l) ? l->name : "(nil)");
			if ( (c = sccp_channel_get_active(d)) ) {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: Trying to put on hold the active call (%d) on line %s\n", d->id, c->callid, (l) ? l->name : "(nil)");
				/* there is an active call, let's put it on hold first */
				if (!sccp_channel_hold(c)) {
					sccp_log(1)(VERBOSE_PREFIX_3 "%s: Hold failed on call (%d), line %s\n", d->id, c->callid, (l) ? l->name : "(nil)");
					return;
				}
			}
			sccp_dev_set_activeline(l);
			sccp_dev_set_cplane(l,1);
			if (!l->channelCount)
				sccp_channel_newcall(l,NULL);
			break;

		case SKINNY_BUTTONTYPE_SPEEDDIAL:
			k = sccp_dev_speed_find_byindex(d, instance);

			if (k) {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: Speeddial Button (%d) pressed, configured number is (%s)\n", d->id, instance, k->ext);
				c = sccp_channel_get_active(d);
				if (c) {
					ast_mutex_lock(&c->lock);
					if ( (c->state == SCCP_CHANNELSTATE_DIALING) || (c->state == SCCP_CHANNELSTATE_OFFHOOK) ) {
						sccp_copy_string(c->owner->exten, k->ext, sizeof(c->owner->exten));
   						c->digittimeout = time(0)+GLOB(digittimeout);
						if (c->state == SCCP_CHANNELSTATE_OFFHOOK)
							sccp_indicate_nolock(c, SCCP_CHANNELSTATE_DIALING);
						ast_mutex_unlock(&c->lock);
						break;
					}
					ast_mutex_unlock(&c->lock);
					sccp_pbx_senddigits(c, k->ext);
				} else {
					// Pull up a channel
					l = s->device->currentLine;
					if (l) {
						sccp_channel_newcall(l, k->ext);
					}
				}
			} else {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: Speeddial Button (%d) pressed, no assigned number\n", d->id, instance);
			}
			break;

		case SKINNY_BUTTONTYPE_HOLD:
			/* this is the hard hold button. When we are here we are putting on hold the active_channel */
			sccp_log(1)(VERBOSE_PREFIX_3 "%s: Hold/Resume Button pressed on line (%d)\n", d->id, instance);
			l = sccp_line_find_byid(d, instance);
			if (!l) {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: No line (%d) found\n", d->id, instance);
				return;
			}

			if ( (c = sccp_channel_find_bystate_on_line(l, SCCP_CHANNELSTATE_CONNECTED)) ) {
				sccp_channel_hold(c);
			} else if ( (c = sccp_channel_find_bystate_on_line(l, SCCP_CHANNELSTATE_HOLD)) ) {
				sccp_channel_resume(c);
			} else {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: No call to resume/hold found on line %d\n", d->id, instance);
			}
			break;

		case SKINNY_BUTTONTYPE_TRANSFER: // Handle transfer via the softkey
			c = sccp_channel_get_active(d);
			if (c)
				sccp_channel_transfer(c);
			break;

		case SKINNY_BUTTONTYPE_VOICEMAIL: // Get a new Line and Dial the Voicemail.
			sccp_log(1)(VERBOSE_PREFIX_3 "%s: Voicemail Button pressed on line (%d)\n", d->id, instance);
			c = sccp_channel_get_active(d);
			if (c) {
				if (!c->line || ast_strlen_zero(c->line->vmnum)) {
					sccp_log(1)(VERBOSE_PREFIX_3 "%s: No voicemail number configured on line %d\n", d->id, instance);
					return;
				}
				ast_mutex_lock(&c->lock);
				if (c->state == SCCP_CHANNELSTATE_OFFHOOK) {
					sccp_copy_string(c->owner->exten, c->line->vmnum, sizeof(c->owner->exten));
					c->digittimeout = time(0)+GLOB(digittimeout);
					sccp_indicate_nolock(c, SCCP_CHANNELSTATE_DIALING);
				} else {
					ast_log(LOG_NOTICE, "Cannot call voicemail while call in progress (%d)\n", c->callid);
				}
				ast_mutex_unlock(&c->lock);
				sccp_dev_check_mwi(d);
				return;
			}
			if (!instance)
				instance = 1;
			l = sccp_line_find_byid(d, instance);
			if (!l) {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: No line (%d) found\n", d->id, instance);
				return;
			}
			if (!ast_strlen_zero(l->vmnum)) {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: Dialing voicemail %s\n", d->id, l->vmnum);
				sccp_channel_newcall(l, l->vmnum);
				sccp_dev_check_mwi(d);
			} else {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: No voicemail number configured on line %d\n", d->id, instance);
			}
			break;

		case SKINNY_BUTTONTYPE_CONFERENCE:
			ast_log(LOG_NOTICE, "%s: Conference Button is not yet handled. working on implementation\n", d->id);
			break;

		case SKINNY_BUTTONTYPE_FORWARDALL: // Call forward all
			ast_log(LOG_NOTICE, "%s: Call forwarding is not yet handled. working on implementation\n", d->id);
			break;

		case SKINNY_BUTTONTYPE_CALLPARK: // Call parking
#ifdef CS_SCCP_PARK
			c = sccp_channel_get_active(d);
			if (!c) {
				sccp_log(1)(VERBOSE_PREFIX_3 "%s: Cannot park while no calls in progress\n", d->id);
				return;
			}
			sccp_channel_park(c);
#else
    		sccp_log(10)(VERBOSE_PREFIX_3 "### Native park was not compiled in\n");
#endif
			break;

		default:
			ast_log(LOG_NOTICE, "%s: Don't know how to deal with stimulus %d with Phonetype %s(%d) \n", d->id, stimulus, skinny_devicetype2str(d->skinny_type), d->skinny_type);
			break;
	}
}

void sccp_handle_offhook(sccp_session_t * s, sccp_moo_t * r) {
	sccp_line_t * l;
	sccp_channel_t * c;
	sccp_device_t * d = s->device;
	if (!d)
		return;

	if ( (c = sccp_channel_get_active(d)) ) {
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Taken Offhook with a call (%d) in progess. Skip it!\n", d->id, c->callid);
		return;
	}

	/* we need this for callwaiting, hold, answer and stuff */
	d->state = SCCP_DEVICESTATE_OFFHOOK;

	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Taken Offhook\n", d->id);

	if (!d->lines) {
    	ast_log(LOG_NOTICE, "No lines registered on %s for take OffHook\n", s->device->id);
    	sccp_dev_displayprompt(s->device, NULL, "No lines registered!", 0);
    	sccp_dev_starttone(s->device, SKINNY_TONE_BEEPBONK);
    	return;
    }

	l = sccp_dev_get_activeline(s->device);

	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Using line %s\n", s->device->id, l->name);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Looking for a ringing channel on line %s\n", s->device->id, l->name);

  /* Check if there is a call comming in on our currently selected line. */

	c = sccp_channel_find_bystate_on_line(l, SCCP_CHANNELSTATE_RINGIN);

	if (c) {
    	/* Answer the ringing channel. */
    	sccp_channel_answer(c);
	} else {
		/* make a new call with no number */
		sccp_channel_newcall(l, NULL);
	}
}

void sccp_handle_onhook(sccp_session_t * s, sccp_moo_t * r) {
	sccp_channel_t * c;
	sccp_line_t * l = NULL;
	sccp_device_t * d = s->device;

	if (!s || !d) {
		ast_log(LOG_NOTICE, "No device to put OnHook\n");
		return;
	}

	/* we need this for callwaiting, hold, answer and stuff */
	d->state = SCCP_DEVICESTATE_ONHOOK;
	sccp_log(1)(VERBOSE_PREFIX_3 "%s is Onhook\n", s->device->id);

	if (!d->lines) {
		ast_log(LOG_NOTICE, "No lines registered on %s to put OnHook\n", s->device->id);
	return;
	}

	/* get the active channel */
	c = sccp_channel_get_active(d);

	if (!c) {
		sccp_dev_set_speaker(d, SKINNY_STATIONSPEAKER_OFF);
		sccp_dev_set_microphone(d, SKINNY_STATIONMIC_OFF);
		sccp_dev_stoptone(d);
		if (!sccp_channel_check_ringback(d))
			sccp_dev_check_displayprompt(d);
		return;
	}

	l = c->line;
	sccp_channel_endcall(c);

	/* we set again the mwi lamp */
	if (d->hasMessages) {
		if (d->ringermode == SKINNY_STATION_RINGOFF) {
			sccp_dev_set_mwi(d, 0, 1);
			sccp_dev_displayprompt(d, NULL, SKINNY_DISP_YOU_HAVE_VOICEMAIL, 10);
		}
		ast_mutex_lock(&l->lock);
		if (l->hasMessages && !l->channels)
			sccp_dev_set_mwi(d, l->instance, 1);
		ast_mutex_unlock(&l->lock);
	}
	return;
}

void sccp_handle_headset(sccp_session_t * s, sccp_moo_t * r) {
  // XXX:T: What should we do here?
}

void sccp_handle_capabilities_res(sccp_session_t * s, sccp_moo_t * r) {
  int i;
  uint8_t codec;
  int astcodec;
  uint8_t n = letohl(r->msg.CapabilitiesResMessage.lel_count);
  s->device->capability = 0;
  sccp_log(10)(VERBOSE_PREFIX_3 "%s: Device has %d Capabilities\n", s->device->id, n);
  for (i = 0 ; i < n; i++) {
    codec = letohl(r->msg.CapabilitiesResMessage.caps[i].lel_payloadCapability);
    astcodec = sccp_codec_skinny2ast(codec);
	s->device->capability |= astcodec;
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: CISCO:%6d %-25s AST:%6d %s\n", s->device->id, codec, skinny_codec2str(codec), astcodec, ast_codec2str(astcodec));
  }
}


void sccp_handle_soft_key_template_req(sccp_session_t * s, sccp_moo_t * r) {
  uint8_t i;
  const uint8_t c = sizeof(softkeysmap);
  sccp_moo_t * r1;

  REQ(r1, SoftKeyTemplateResMessage);
  r1->msg.SoftKeyTemplateResMessage.lel_softKeyOffset = htolel(0);

  for (i = 0; i < c; i++) {
    sccp_log(10)(VERBOSE_PREFIX_3 "%s: Button(%d)[%2d] = %s\n", s->device->id, i, i+1, skinny_lbl2str(softkeysmap[i]));
	r1->msg.SoftKeyTemplateResMessage.definition[i].softKeyLabel[0] = 128;
	r1->msg.SoftKeyTemplateResMessage.definition[i].softKeyLabel[1] = softkeysmap[i];
    r1->msg.SoftKeyTemplateResMessage.definition[i].lel_softKeyEvent = htolel(i+1);
  }

  r1->msg.SoftKeyTemplateResMessage.lel_softKeyCount = htolel(c);
  r1->msg.SoftKeyTemplateResMessage.lel_totalSoftKeyCount = htolel(c);
  sccp_dev_send(s->device, r1);
}


void sccp_handle_soft_key_set_req(sccp_session_t * s, sccp_moo_t * r) {
  sccp_device_t * d = s->device;
  const softkey_modes * v = SoftKeyModes;
  const	uint8_t	v_count = ( sizeof(SoftKeyModes)/sizeof(softkey_modes) );
  int iKeySetCount = 0;
  sccp_moo_t * r1;
  uint8_t i = 0;

  REQ(r1, SoftKeySetResMessage);
  r1->msg.SoftKeySetResMessage.lel_softKeySetOffset = htolel(0);

  sccp_log(10)(VERBOSE_PREFIX_3 "%s: Transfer is  %s\n", d->id, (d->transfer) ? "enabled" : "disabled");

  for (i = 0; i < v_count; i++) {
    const uint8_t * b = v->ptr;
    uint8_t c, i = 0;

    sccp_log(10)(VERBOSE_PREFIX_3 "%s: Set[%-2d]= ", d->id, v->id);

    for ( c = 0; c < v->count; c++) {
      /* look for the SKINNY_LBL_ number in the softkeysmap */
      if ( (b[c] == SKINNY_LBL_TRANSFER) && (!d->transfer) ) {
      	continue;
	  }
      for (i = 0; i < sizeof(softkeysmap); i++) {
		if (b[c] == softkeysmap[i]) {
			sccp_log(10)("%-2d:%-10s ", c, skinny_lbl2str(softkeysmap[i]));
			r1->msg.SoftKeySetResMessage.definition[v->id].softKeyTemplateIndex[c] = (i+1);
			break;
		}
	  }
    }

    sccp_log(10)("\n");
    v++;
    iKeySetCount++;
  };

  sccp_log(10)( VERBOSE_PREFIX_3 "There are %d SoftKeySets.\n", iKeySetCount);

  r1->msg.SoftKeySetResMessage.lel_softKeySetCount = htolel(iKeySetCount);
  r1->msg.SoftKeySetResMessage.lel_totalSoftKeySetCount = htolel(iKeySetCount); // <<-- for now, but should be: iTotalKeySetCount;

  sccp_dev_send(s->device, r1);
  sccp_dev_set_keyset(s->device, 0, 0, KEYMODE_ONHOOK);
}

void sccp_handle_time_date_req(sccp_session_t * s, sccp_moo_t * req) {
  time_t timer = 0;
  struct tm * cmtime = NULL;
  sccp_moo_t * r1;
  REQ(r1, DefineTimeDate);

  if (NULL == s) {
       ast_log(LOG_WARNING,"Session no longer valid\n");
       return;
  }

  /* modulate the timezone by full hours only */
  timer = time(0) + (s->device->tz_offset * 3600);
  cmtime = localtime(&timer);

  r1->msg.DefineTimeDate.lel_year = htolel(cmtime->tm_year+1900);
  r1->msg.DefineTimeDate.lel_month = htolel(cmtime->tm_mon+1);
  r1->msg.DefineTimeDate.lel_dayOfWeek = htolel(cmtime->tm_wday);
  r1->msg.DefineTimeDate.lel_day = htolel(cmtime->tm_mday);
  r1->msg.DefineTimeDate.lel_hour = htolel(cmtime->tm_hour);
  r1->msg.DefineTimeDate.lel_minute = htolel(cmtime->tm_min);
  r1->msg.DefineTimeDate.lel_seconds = htolel(cmtime->tm_sec);
  r1->msg.DefineTimeDate.lel_milliseconds = htolel(0);
  r1->msg.DefineTimeDate.lel_systemTime = htolel(timer);
  sccp_dev_send(s->device, r1);
  sccp_log(10)(VERBOSE_PREFIX_3 "%s: Send date/time\n", s->device->id);
}

void sccp_handle_keypad_button(sccp_session_t * s, sccp_moo_t * r) {
	sccp_channel_t * c = sccp_channel_get_active(s->device);
	int event = letohl(r->msg.KeypadButtonMessage.lel_kpButton);
	char resp = '\0';
	int len = 0;

	if (!c) {
		ast_log(LOG_NOTICE, "Device %s sent a Keypress, but there is no active channel!\n", s->device->id);
		return;
	}

	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Cisco Digit: %08x (%d) on line %s\n", c->device->id, event, event, c->line->name);

	if (event < 10)
		resp = '0' + event;
	else if (event == 14)
		resp = '*';
	else if (event == 15)
		resp = '#';

	ast_mutex_lock(&c->lock);

	if (c->state == SCCP_CHANNELSTATE_CONNECTED) {
		/* we have to unlock 'cause the senddigit lock the channel */
		ast_mutex_unlock(&c->lock);
    	sccp_pbx_senddigit(c, resp);
    	return;
	}

	if ( (c->state == SCCP_CHANNELSTATE_DIALING) || (c->state == SCCP_CHANNELSTATE_OFFHOOK) ) {
		len = strlen(c->owner->exten);
		if (len >= (AST_MAX_EXTENSION - 1) ) {
			sccp_dev_displayprompt(NULL,c, "No more digits", 5);
		} else {
		c->owner->exten[len++] = resp;
    	c->owner->exten[len] = '\0';
    	c->digittimeout = time(0)+GLOB(digittimeout);
		}
	}
	if (len == 1)
		sccp_indicate_nolock(c, SCCP_CHANNELSTATE_DIALING);
	ast_mutex_unlock(&c->lock);
}


void sccp_handle_soft_key_event(sccp_session_t * s, sccp_moo_t * r) {
	sccp_device_t * d = s->device;
	sccp_channel_t * c = NULL;
	sccp_line_t * l = NULL;
	uint32_t event = letohl(r->msg.SoftKeyEventMessage.lel_softKeyEvent);
	uint32_t line = letohl(r->msg.SoftKeyEventMessage.lel_lineInstance);
	uint32_t callid = letohl(r->msg.SoftKeyEventMessage.lel_callReference);

	if (!d)
		return;

	if ( event > sizeof(softkeysmap) )
		ast_log(LOG_ERROR, "%s: Out of range for Softkey: %s (%d) line=%d callid=%d\n", d->id, skinny_lbl2str(event), event, line, callid);
	event = softkeysmap[event-1];

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Got Softkey: %s (%d) line=%d callid=%d\n", d->id, skinny_lbl2str(event), event, line, callid);

	if (line)
		l = sccp_line_find_byid(s->device, line);

	if (line && callid)
		c = sccp_channel_find_byid_on_line(l, callid);

	switch (event) {
	case SKINNY_LBL_REDIAL:
		sccp_sk_redial(d, l, c);
		break;
	case SKINNY_LBL_NEWCALL:
		sccp_sk_newcall(d, l, c);
		break;
	case SKINNY_LBL_HOLD:
		sccp_sk_hold(d, l, c);
		break;
	case SKINNY_LBL_TRANSFER:
		sccp_sk_transfer(d, l, c);
		break;
	case SKINNY_LBL_CFWDALL:
		sccp_sk_cfwdall(d, l, c);
		break;
	case SKINNY_LBL_CFWDBUSY:
		sccp_sk_cfwdbusy(d, l, c);
		break;
	case SKINNY_LBL_CFWDNOANSWER:
		sccp_sk_cfwdnoanswer(d, l, c);
		break;
	case SKINNY_LBL_BACKSPACE:
		sccp_sk_backspace(d, l, c);
		break;
	case SKINNY_LBL_ENDCALL:
		sccp_sk_endcall(d, l, c);
		break;
	case SKINNY_LBL_RESUME:
		sccp_sk_resume(d, l, c);
		break;
	case SKINNY_LBL_ANSWER:
		sccp_sk_answer(d, l, c);
		break;
	case SKINNY_LBL_PARK:
		sccp_sk_park(d, l, c);
		break;
	case SKINNY_LBL_TRNSFVM:
		sccp_sk_trnsfvm(d, l, c);
		break;
	case SKINNY_LBL_DND:
		sccp_sk_dnd(d, l, c);
		break;
	case SKINNY_LBL_DIRTRFR:
		sccp_sk_dirtrfr(d, l, c);
		break;
	default:
		ast_log(LOG_WARNING, "Don't know how to handle keypress %d\n", event);
	}
}

void sccp_handle_open_receive_channel_ack(sccp_session_t * s, sccp_moo_t * r) {
	struct sockaddr_in sin;
	char iabuf[INET_ADDRSTRLEN];
	sccp_channel_t * c;
	sccp_device_t * d;

	if (!s)
		return;

	d = s->device;
	
	sin.sin_family = AF_INET;
	memcpy(&sin.sin_addr, &r->msg.OpenReceiveChannelAck.bel_ipAddr, sizeof(sin.sin_addr));
	sin.sin_port = htons(htolel(r->msg.OpenReceiveChannelAck.lel_portNumber));

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Got OpenChannel ACK.  Status: %d, RemoteIP: %s, Port: %d, PassThruId: %d\n",
		(d) ? d->id : "SCCP",
		letohl(r->msg.OpenReceiveChannelAck.lel_orcStatus),
		ast_inet_ntoa(iabuf, sizeof(iabuf), sin.sin_addr),
		ntohs(sin.sin_port),
		letohl(r->msg.OpenReceiveChannelAck.lel_passThruPartyId));

	c = sccp_channel_find_bypartyid(letohl(r->msg.OpenReceiveChannelAck.lel_passThruPartyId));
	if (c) {
		memcpy(&c->rtp_addr, &sin, sizeof(sin));
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: open RTP media at %s:%d\n", s->device->id,ast_inet_ntoa(iabuf, sizeof(iabuf), c->rtp_addr.sin_addr), ntohs(c->rtp_addr.sin_port));
		sccp_channel_startmediatransmission(c);
	} else {
		ast_log(LOG_ERROR, "%s: No channel with this PassThruId!\n",(d) ? d->id : "SCCP" );
	}
}

void sccp_handle_version(sccp_session_t * s, sccp_moo_t * r) {
	sccp_moo_t * r1;
	char * verStr = sccp_helper_getversionfor(s);

	REQ(r1, VersionMessage);
	sccp_copy_string(r1->msg.VersionMessage.requiredVersion, verStr, sizeof(r1->msg.VersionMessage.requiredVersion));
	sccp_dev_send(s->device, r1);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Sending version number\n", s->device->id);
}

void sccp_handle_ConnectionStatistics(sccp_session_t * s, sccp_moo_t * r) {
	ast_verbose(VERBOSE_PREFIX_3 "%s: Statistics from %s callid: %d Packets sent: %d rcvd: %d lost: %d jitter: %d latency: %d\n", s->device->id, r->msg.ConnectionStatisticsRes.DirectoryNumber,
		letohl(r->msg.ConnectionStatisticsRes.lel_CallIdentifier),
		letohl(r->msg.ConnectionStatisticsRes.lel_SentPackets),
		letohl(r->msg.ConnectionStatisticsRes.lel_RecvdPackets),
		letohl(r->msg.ConnectionStatisticsRes.lel_LostPkts),
		letohl(r->msg.ConnectionStatisticsRes.lel_Jitter),
		letohl(r->msg.ConnectionStatisticsRes.lel_latency)
		);
}

void sccp_handle_ServerResMessage(sccp_session_t * s, sccp_moo_t * r) {
	sccp_moo_t * r1;
	struct ast_hostent	ahp;
	char iabuf[INET_ADDRSTRLEN];
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Sending servers message\n", s->device->id);
	if (ast_gethostbyname(ast_inet_ntoa(iabuf, sizeof(iabuf), GLOB(bindaddr).sin_addr), &ahp)) {
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: My dns name %s\n", s->device->id, ahp.hp.h_addr);
	}

	REQ(r1, ServerResMessage);
	sccp_copy_string(r1->msg.ServerResMessage.server[0].serverName, ahp.hp.h_name, sizeof(r1->msg.ServerResMessage.server[0].serverName));
	r1->msg.ServerResMessage.serverListenPort[0] = GLOB(ourport);
	r1->msg.ServerResMessage.serverIpAddr[0] = s->ourip.s_addr;
	sccp_dev_send(s->device, r1);
}

void sccp_handle_ConfigStatMessage(sccp_session_t * s, sccp_moo_t * r) {
	sccp_moo_t * r1;
	uint8_t lines = 0;
	uint8_t speeddials = 0;
	sccp_device_t * d;
	sccp_line_t * l;
	sccp_speed_t * k;

	if (!s)
		return;
	d = s->device;
	if (!d)
		return;
	ast_mutex_lock(&d->lock);
	ast_mutex_lock(&GLOB(lines_lock));
	l = d->lines;
	while (l) {
		lines++;
		l = l->next_on_device;
	}
	k = d->speed_dials;
	while (k) {
		speeddials++;
		k = k->next;
	}
	ast_mutex_unlock(&GLOB(lines_lock));
	ast_mutex_unlock(&d->lock);


	REQ(r1, ConfigStatMessage);
	sccp_copy_string(r1->msg.ConfigStatMessage.deviceName, s->device->id, sizeof(r1->msg.ConfigStatMessage.deviceName));
	r1->msg.ConfigStatMessage.lel_stationInstance = htolel(1);
	r1->msg.ConfigStatMessage.lel_numberLines	     = htolel(lines);
	r1->msg.ConfigStatMessage.lel_numberSpeedDials 	     = htolel(speeddials);
	sccp_dev_send(s->device, r1);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Sending ConfigStatMessage, lines %d, speedials %d\n", d->id, lines, speeddials);
}

void sccp_handle_EnblocCallMessage(sccp_session_t * s, sccp_moo_t * r) {
	ast_log(LOG_NOTICE, "EnblocCallMessage recognized - not implemented yet\n");
}

void sccp_handle_forward_stat_req(sccp_session_t * s, sccp_moo_t * r) {
	sccp_device_t * d = s->device;
	sccp_moo_t * r1 = NULL;

	if (!d)
		return;

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Got Forward Status Request.  Line: %d\n", d->id, letohl(r->msg.ForwardStatReqMessage.lel_lineNumber));

	REQ(r1, ForwardStatMessage);
	r1->msg.ForwardStatReqMessage.lel_lineNumber = r->msg.ForwardStatReqMessage.lel_lineNumber;
	/* XXX add support */
/*	r1->msg.ForwardStatMessage.lel_status = htolel(1); */ /* 0 inactive 1 active */
/*	r1->msg.ForwardStatMessage.lel_lineNumber = htolel(1);
	r1->msg.ForwardStatMessage.lel_cfwdallstatus = htolel(1);
	sccp_copy_string(r1->msg.ForwardStatMessage.cfwdallnumber, "0", sizeof(r1->msg.ForwardStatMessage.cfwdallnumber));
*/
	sccp_dev_send(s->device, r1);
}
