/*
 * (SCCP*)
 *
 * An implementation of Skinny Client Control Protocol (SCCP)
 *
 * Sergio Chersovani (mlists@c-net.it)
 *
 * Reworked, but based on chan_sccp code.
 * The original chan_sccp driver that was made by Zozo which itself was derived from the chan_skinny driver.
 * Modified by Jan Czmok and Julien Goodwin
 *
 * This program is free software and may be modified and
 * distributed under the terms of the GNU Public License.
 */

#include "config.h"
#include "chan_sccp.h"
#include "sccp_actions.h"
#include "sccp_utils.h"
#include "sccp_device.h"
#include "sccp_pbx.h"
#include "sccp_channel.h"
#include "sccp_indicate.h"
#include <asterisk/pbx.h>
#include <asterisk/callerid.h>
#include <asterisk/acl.h>
#include <asterisk/musiconhold.h>
#ifdef CS_SCCP_PARK
#include <asterisk/features.h>
#endif

static uint32_t callCount = 1;
AST_MUTEX_DEFINE_STATIC(callCountLock);

sccp_channel_t * sccp_channel_allocate(sccp_line_t * l) {
/* this just allocate a sccp channel (not the asterisk channel, for that look at sccp_pbx_channel_allocate) */
	sccp_channel_t * c;
	sccp_device_t * d;

  /* If there is no current line, then we can't make a call in, or out. */
	if (!l) {
		ast_log(LOG_ERROR, "SCCP: Tried to open channel on a device with no lines\n");
	return NULL;
	}

	d = l->device;

	if (!d->session) {
		ast_log(LOG_ERROR, "SCCP: Tried to open channel on device %s without a session\n", d->id);
		return NULL;
	}

	c = malloc(sizeof(sccp_channel_t));
	if (!c) {
		/* error allocating memory */
		ast_log(LOG_ERROR, "%s: No memory to allocate channel on line %s\n",d->id, l->name);
		return NULL;
	}
	memset(c, 0, sizeof(sccp_channel_t));
	ast_mutex_init(&c->lock);

	/* default ringermode SKINNY_STATION_OUTSIDERING. Change it with SCCPRingerMode app */
	c->ringermode = SKINNY_STATION_OUTSIDERING;
	/* inbound for now. It will be changed later on outgoing calls */
	c->calltype = SKINNY_CALLTYPE_INBOUND;

	ast_mutex_lock(&callCountLock);
	c->callid = callCount++;
	ast_mutex_unlock(&callCountLock);

	c->line = l;
	c->device = d;

	ast_mutex_lock(&l->lock);
	l->channelCount++;
	ast_mutex_unlock(&l->lock);

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: New channel number: %d on line %s\n", d->id, c->callid, l->name);

	/* put it on the top of the lists */

	ast_mutex_lock(&GLOB(channels_lock));
	c->next = GLOB(channels);
	if (GLOB(channels)) /* first one */
		GLOB(channels)->prev = c;
	GLOB(channels) = c;
	ast_mutex_unlock(&GLOB(channels_lock));

	c->next_on_line = l->channels;
	if (l->channels) /* first one */
		l->channels->prev_on_line = c;
	l->channels = c;

	return c;
}

sccp_channel_t * sccp_channel_get_active(sccp_device_t * d) {
	sccp_channel_t * c;
	if (!d)
		return NULL;
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: getting the active channel on device\n",d->id);
	ast_mutex_lock(&d->lock);
	c = d->active_channel;
	ast_mutex_unlock(&d->lock);
	/* go safe! Will remove it soon after test the active_channel XXX*/
/*	if (c)
		return c;
	if ( !(c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_CONNECTED)) )
		if ( !(c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_RINGOUT)) )
			if ( !(c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_OFFHOOK)) )
				if ( !(c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_BUSY)) )
					c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_PROCEED);
*/
	return c;
}

void sccp_channel_set_active(sccp_channel_t * c) {
	sccp_device_t * d = c->device;
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Set the active channel %d on device\n", d->id, (c) ? c->callid : 0);
	ast_mutex_lock(&d->lock);
	d->active_channel = c;
	d->currentLine = c->line;
	ast_mutex_unlock(&d->lock);
}

void sccp_channel_send_callinfo(sccp_channel_t * c) {
	sccp_moo_t * r;

	if (!c->owner) {
		ast_log(LOG_ERROR, "%s: Channel %d doesn't have an owner!\n", c->line->device->id, c->callid);
		return;
	}

	REQ(r, CallInfoMessage);

	if (c->callingPartyName)
		sccp_copy_string(r->msg.CallInfoMessage.callingPartyName, c->callingPartyName, sizeof(r->msg.CallInfoMessage.callingPartyName));
	if (c->callingPartyNumber)
		sccp_copy_string(r->msg.CallInfoMessage.callingParty, c->callingPartyNumber, sizeof(r->msg.CallInfoMessage.callingParty));

	if (c->calledPartyName)
		sccp_copy_string(r->msg.CallInfoMessage.calledPartyName, c->calledPartyName, sizeof(r->msg.CallInfoMessage.calledPartyName));
	if (c->calledPartyNumber)
		sccp_copy_string(r->msg.CallInfoMessage.calledParty, c->calledPartyNumber, sizeof(r->msg.CallInfoMessage.calledParty));

	r->msg.CallInfoMessage.lel_lineId   = htolel(c->line->instance);
	r->msg.CallInfoMessage.lel_callRef  = htolel(c->callid);
	r->msg.CallInfoMessage.lel_callType = htolel(c->calltype);
	sccp_dev_send(c->line->device, r);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Send callinfo for channel %d\n", c->line->device->id, c->callid);
}

// XXX:T: whoops - need to move to device.
void sccp_channel_set_callstate(sccp_channel_t * c, uint8_t state) {
	sccp_moo_t * r;
	c->callstate = state;

	if (c->device->skinny_type < 6 || (c->device->skinny_type > 8 && c->device->skinny_type <= 255)) return;   /* only for telecaster */

	REQ(r, CallStateMessage);
	r->msg.CallStateMessage.lel_callState = htolel(state);
	r->msg.CallStateMessage.lel_lineInstance	= htolel((c) ? c->line->instance : 0);
	r->msg.CallStateMessage.lel_callReference = htolel((c) ? c->callid : 0);
	sccp_dev_send(c->device, r);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Send and Set the call state %s(%d) for %s-%d\n",
	c->device->id, sccp_callstate2str(state), state, ((c) ? c->line->name : "0"), ((c) ? c->callid : 0));
}

/* temporary function for hints */
void sccp_channel_set_callstate1(sccp_device_t * d, uint8_t instance, uint32_t callid, uint8_t state) {
	sccp_moo_t * r;

	if (d->skinny_type < 6 || (d->skinny_type > 8 && d->skinny_type <= 255)) return;   /* only for telecaster */

	REQ(r, CallStateMessage);
	r->msg.CallStateMessage.lel_callState = htolel(state);
	r->msg.CallStateMessage.lel_lineInstance	= htolel(instance);
	r->msg.CallStateMessage.lel_callReference = htolel(callid);
	r->msg.CallStateMessage.lel_unknown1 = htolel(0);
//	r->msg.CallStateMessage.lel_unknown2 = htolel(0);
	sccp_dev_send(d, r);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Send and Set the call state1 %s(%d)\n",
	d->id, sccp_callstate2str(state), state);
}

void sccp_channel_set_callingparty(sccp_channel_t * c, char *name, char *number) {
	sccp_device_t * d;
	if (!c)
		return;

	d = c->device;

	if (name && strncmp(name, c->callingPartyName, StationMaxNameSize - 1)) {
		sccp_copy_string(c->callingPartyName, name, sizeof(c->callingPartyName));
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Set callingParty Name %s on channel %d\n",d->id, c->callingPartyName, c->callid);
	}

	if (number && strncmp(number, c->callingPartyNumber, StationMaxDirnumSize - 1)) {
		sccp_copy_string(c->callingPartyNumber, number, sizeof(c->callingPartyNumber));
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Set callingParty Number %s on channel %d\n", d->id, c->callingPartyNumber, c->callid);
	}

	return;
}

void sccp_channel_set_calledparty(sccp_channel_t * c, char *name, char *number) {
	sccp_device_t * d;

	if (!c)
		return;

	d = c->device;

	if (name && strncmp(name, c->calledPartyName, StationMaxNameSize - 1)) {
		sccp_copy_string(c->calledPartyName, name, sizeof(c->calledPartyNumber));
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Set calledParty Name %s on channel %d\n", d->id, c->calledPartyName, c->callid);
	}

	if (number && strncmp(number, c->calledPartyNumber, StationMaxDirnumSize - 1)) {
		sccp_copy_string(c->calledPartyNumber, number, sizeof(c->callingPartyNumber));
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Set calledParty Number %s on channel %d\n", d->id, c->calledPartyNumber, c->callid);
	}
}

void sccp_channel_StatisticsRequest(sccp_channel_t * c) {
	sccp_moo_t * r;

	if (!c)
		return;

	REQ(r, ConnectionStatisticsReq);

	/* XXX need to test what we have to copy in the DirectoryNumber */
	if (c->calltype == SKINNY_CALLTYPE_OUTBOUND)
		sccp_copy_string(r->msg.ConnectionStatisticsReq.DirectoryNumber, c->calledPartyNumber, sizeof(r->msg.ConnectionStatisticsReq.DirectoryNumber));
	else
		sccp_copy_string(r->msg.ConnectionStatisticsReq.DirectoryNumber, c->callingPartyNumber, sizeof(r->msg.ConnectionStatisticsReq.DirectoryNumber));

	r->msg.ConnectionStatisticsReq.lel_callReference = htolel((c) ? c->callid : 0);
	r->msg.ConnectionStatisticsReq.lel_StatsProcessing = htolel(SKINNY_STATSPROCESSING_CLEAR);
	sccp_dev_send(c->line->device, r);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Requesting CallStatisticsAndClear from Phone\n", c->line->device->id);
}

void sccp_channel_openreceivechannel(sccp_channel_t * c) {
	sccp_moo_t * r;
	int payloadType = sccp_codec_ast2skinny(c->owner->readformat);

	REQ(r, OpenReceiveChannel);
	r->msg.OpenReceiveChannel.lel_conferenceId = htolel(c->callid);
	r->msg.OpenReceiveChannel.lel_passThruPartyId = htolel((uint32_t)c) ;
	r->msg.OpenReceiveChannel.lel_millisecondPacketSize = htolel(20);
	r->msg.OpenReceiveChannel.lel_payloadType = htolel((payloadType) ? payloadType : 4);
	r->msg.OpenReceiveChannel.qualifierIn.lel_vadValue = htolel(c->line->echocancel);
	sccp_dev_send(c->line->device, r);
	/* create the rtp stuff. It must be create before setting the channel AST_STATE_UP. otherwise no audio will be played */
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Send openreceivechannel on channel %d\n", c->line->device->id, c->callid);
		if (!c->rtp) {
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Starting RTP on channel %s-%d\n", c->device->id, c->line->name, c->callid);
		sccp_channel_start_rtp(c);
	}
	if (!c->rtp) {
		ast_log(LOG_WARNING, "%s: Error opening RTP for channel %s-%d\n", c->device->id, c->line->name, c->callid);
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONGESTION);
		return;
	}

}

void sccp_channel_startmediatransmission(sccp_channel_t * c) {
	sccp_moo_t * r;
	struct sockaddr_in sin;
	char iabuf[INET_ADDRSTRLEN];
	int payloadType = sccp_codec_ast2skinny(c->owner->readformat);

	ast_rtp_set_peer(c->rtp, &c->rtp_addr);
	ast_rtp_get_us(c->rtp, &sin);
	ast_rtp_settos(c->rtp, c->line->rtptos);

	REQ(r, StartMediaTransmission);
	r->msg.StartMediaTransmission.lel_conferenceId = htolel(c->callid);
	r->msg.StartMediaTransmission.lel_passThruPartyId = htolel((uint32_t)c);
	/* r->msg.StartMediaTransmission.bel_remoteIpAddr = */
	memcpy(&r->msg.StartMediaTransmission.bel_remoteIpAddr, &sin.sin_addr, 4);
	r->msg.StartMediaTransmission.lel_remotePortNumber = htolel(ntohs(sin.sin_port));
	r->msg.StartMediaTransmission.lel_millisecondPacketSize = htolel(20);
	r->msg.StartMediaTransmission.lel_payloadType = htolel((payloadType) ? payloadType : 4);
	r->msg.StartMediaTransmission.qualifierOut.lel_precedenceValue = htolel(c->line->rtptos);
	r->msg.StartMediaTransmission.qualifierOut.lel_ssValue = htolel(c->line->silencesuppression); // Silence supression
	r->msg.StartMediaTransmission.qualifierOut.lel_maxFramesPerPacket = 0;
	sccp_dev_send(c->line->device, r);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Start media RTP on %s:%d with tos %d\n",c->line->device->id, ast_inet_ntoa(iabuf, sizeof(iabuf), sin.sin_addr), ntohs(sin.sin_port), c->line->rtptos);
}


void sccp_channel_closereceivechannel(sccp_channel_t * c) {
	sccp_moo_t * r;
	REQ(r, CloseReceiveChannel);
	r->msg.CloseReceiveChannel.lel_conferenceId = htolel(c ? c->callid : 0);
	r->msg.CloseReceiveChannel.lel_passThruPartyId = htolel((uint32_t)c);
	sccp_dev_send(c->line->device, r);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Close openreceivechannel on channel %d\n",c->line->device->id, c->callid);
	sccp_channel_stopmediatransmission(c);
}

void sccp_channel_stopmediatransmission(sccp_channel_t * c) {
	sccp_moo_t * r;
	if (c->rtp) {
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: Stopping RTP on channel %s-%d\n", c->device->id, c->line->name, c->callid);
		sccp_channel_stop_rtp(c);
	}
	REQ(r, StopMediaTransmission);
	r->msg.CloseReceiveChannel.lel_conferenceId = htolel(c ? c->callid : 0);
	r->msg.CloseReceiveChannel.lel_passThruPartyId = htolel((uint32_t)c);
	sccp_dev_send(c->line->device, r);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Stop media transmission on channel %d\n",c->line->device->id, c->callid);
	/* requesting statistics */
	sccp_channel_StatisticsRequest(c);
}

int sccp_channel_check_ringback(sccp_device_t * d) {
	sccp_channel_t * c;
	ast_mutex_lock(&d->lock);
	if (d->state == SCCP_DEVICESTATE_OFFHOOK) {
		ast_mutex_unlock(&d->lock);
		return 0;
	}
	ast_mutex_unlock(&d->lock);
	/* we are on hook let's ring again the call transfer or the waiting call*/
	c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_CALLTRANSFER);
	if (!c)
		/* we have stopped the ringer tone so let's ring it again */
		c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_RINGIN);
	if (!c)
		c = sccp_channel_find_bystate_on_device(d, SCCP_CHANNELSTATE_CALLWAITING);

	if (c) {
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_RINGIN);
		return 1;
	}
	return 0;
}

void sccp_channel_endcall(sccp_channel_t * c) {
	sccp_device_t * d;

	if (!c)
		return;

	/* this is a station active endcall or onhook */
	d = c->device;
	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Ending call %d on line %s\n", d->id, c->callid, c->line->name);

	ast_mutex_lock(&c->lock);
	if (c->owner) {
		CS_AST_CHANNEL_PVT(c->owner) = NULL;
		// sccp_ast_setstate(c, AST_STATE_DOWN);
		ast_softhangup(c->owner, AST_SOFTHANGUP_EXPLICIT);
	}
	if (c->rtp) {
		sccp_channel_closereceivechannel(c);
		sccp_channel_stop_rtp(c);
	}

	if (c->state == SCCP_CHANNELSTATE_RINGIN) {
		sccp_indicate_nolock(c, SCCP_CHANNELSTATE_HANGUP_RINGIN);
	} else if (c->state == SCCP_CHANNELSTATE_HOLD) {
		sccp_indicate_nolock(c, SCCP_CHANNELSTATE_HANGUP_HOLD);
	} else if (c->state == SCCP_CHANNELSTATE_CALLWAITING) {
		sccp_indicate_nolock(c, SCCP_CHANNELSTATE_HANGUP_CALLWAITING);
	} else
		sccp_indicate_nolock(c, SCCP_CHANNELSTATE_ONHOOK);

	ast_mutex_unlock(&c->lock);
	ast_mutex_lock(&GLOB(channels_lock));
	sccp_channel_delete(c);
	ast_mutex_unlock(&GLOB(channels_lock));
	/* any call for a ringback ? */
	if (!sccp_channel_check_ringback(d))
		sccp_dev_check_displayprompt(d);
	return;
}

void sccp_channel_newcall(sccp_line_t * l, char * dial) {
	/* handle outgoing calls */
	sccp_channel_t * c;
	sccp_device_t * d;
	pthread_attr_t attr;
	pthread_t t;

	if (!l) {
		ast_log(LOG_ERROR, "SCCP: Can't allocate SCCP channel if line is not defined!\n");
		return;
	}

	d = l->device;

	/* look if we have a call to put on hold */
	if ( (c = sccp_channel_get_active(d)) ) {
		/* there is an active call, let's put it on hold first */
		if (!sccp_channel_hold(c))
			return;
	}

	c = sccp_channel_allocate(l);

	if (!c) {
		ast_log(LOG_ERROR, "%s: Can't allocate SCCP channel for line %s\n",d->id, l->name);
		return;
	}
	
	/* let's see if the number exists */
	if (dial && !ast_exists_extension(NULL, l->context, dial, 1, l->cid_num)) {
		ast_log(LOG_WARNING, "%s: No extension to dial (%s)\n",l->device->id, dial);
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONGESTION);
		/* wait the onhook to free the SCCP channel */
		return;
	}

	/* ok the number exist. allocate the asterisk channel */
	if (!sccp_pbx_channel_allocate(c)) {
		ast_log(LOG_WARNING, "%s: Unable to allocate a new channel for line %s\n", d->id, l->name);
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONGESTION);
		return;
	}

	/* copy the number to dial in the ast->exten */
	if (dial)
		sccp_copy_string(c->owner->exten, dial, sizeof(c->owner->exten));
	sccp_channel_set_active(c);
	sccp_indicate_lock(c, SCCP_CHANNELSTATE_OFFHOOK);

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  /* let's call it */
	if (ast_pthread_create(&t, &attr, sccp_pbx_startchannel, c->owner)) {
		ast_log(LOG_WARNING, "%s: Unable to create switch thread for channel (%s-%d) %s\n", d->id, l->name, c->callid, strerror(errno));
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONGESTION);
		return;
	}
}

void sccp_channel_answer(sccp_channel_t * c) {
	sccp_line_t * l;
	sccp_device_t * d;
	sccp_channel_t * hold;
	struct ast_channel * bridged;

	if (!c)
		return;
	l = c->line;
	d = c->line->device;
	/* answering an incoming call */
	/* look if we have a call to put on hold */
	if ( (hold = sccp_channel_get_active(d)) ) {
		/* there is an active call, let's put it on hold first */
		if (!sccp_channel_hold(hold))
			return;
	}

	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Answer the channel %s-%d\n", d->id, l->name, c->callid);
	sccp_channel_set_active(c);
    sccp_dev_set_activeline(c->line);
	if (c->owner) {
		/* the old channel state could be CALLTRANSFER, so the bridged channel is on hold */
		bridged = CS_AST_BRIDGED_CHANNEL(c->owner);
		if (bridged && ast_test_flag(bridged, AST_FLAG_MOH)) {
			ast_moh_stop(bridged);
#ifndef CS_AST_HAS_FLAG_MOH
			ast_clear_flag(bridged, AST_FLAG_MOH);
#endif
		}
	}
	sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONNECTED);
	ast_queue_control(c->owner, AST_CONTROL_ANSWER);
}

int sccp_channel_hold(sccp_channel_t * c) {
	sccp_line_t * l;
	sccp_device_t * d;
	struct ast_channel * peer;

	if (!c)
		return 0;
	l = c->line;
	d = c->line->device;

	if (c->state == SCCP_CHANNELSTATE_HOLD)
		return 0;
	/* put on hold an active call */
	if (sccp_channel_get_active(d) != c || c->state != SCCP_CHANNELSTATE_CONNECTED) {
		/* something wrong on the code let's notify it for a fix */
		ast_log(LOG_ERROR, "%s can't put on hold an inactive channel %s-%d\n", d->id, l->name, c->callid);
		/* hard button phones need it */
		sccp_dev_displayprompt(NULL, c, "No active call to put on hold.",5);
		return 0;
	}

	peer = CS_AST_BRIDGED_CHANNEL(c->owner);

	if (!peer) {
		ast_log(LOG_ERROR, "%s can't put on hold this type of channel %s-%d\n", d->id, l->name, c->callid);
		sccp_dev_displayprompt(NULL, c, "Can't put on hold.",5);
		return 0;
	}

    ast_moh_start(peer, NULL);
#ifndef CS_AST_HAS_FLAG_MOH
	ast_set_flag(peer, AST_FLAG_MOH);
#endif
	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Hold the channel %s-%d\n", d->id, l->name, c->callid);
	ast_mutex_lock(&d->lock);
	d->active_channel = NULL;
	ast_mutex_unlock(&d->lock);
	sccp_indicate_lock(c, SCCP_CHANNELSTATE_HOLD);
#ifdef CS_AST_CONTROL_HOLD
	ast_queue_control(c->owner, AST_CONTROL_HOLD);
#endif
	return 1;
}

int sccp_channel_resume(sccp_channel_t * c) {
	sccp_line_t * l;
	sccp_device_t * d;
	sccp_channel_t * hold;

	if (!c)
		return 0;

	l = c->line;
	d = c->line->device;

	/* look if we have a call to put on hold */
	if ( (hold = sccp_channel_get_active(d)) ) {
		/* there is an active call, let's put it on hold first */
		if (!sccp_channel_hold(hold))
			return 0;
	}

	/* resume an active call */
	if (c->state != SCCP_CHANNELSTATE_HOLD && c->state != SCCP_CHANNELSTATE_CALLTRANSFER) {
		/* something wrong on the code let's notify it for a fix */
		ast_log(LOG_ERROR, "%s can't resume the channel %s-%d. Not on hold\n", d->id, l->name, c->callid);
		sccp_dev_displayprompt(NULL, c, "No active call to put on hold",5);
		return 0;
	}

	/* check if we are in the middle of a transfer */
	ast_mutex_lock(&d->lock);
	if (d->transfer_channel == c)
		d->transfer_channel = NULL;
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Transfer aborted on the channel %s-%d\n", d->id, l->name, c->callid);
	ast_mutex_unlock(&d->lock);

    if (CS_AST_BRIDGED_CHANNEL(c->owner)) {
            ast_moh_stop(CS_AST_BRIDGED_CHANNEL(c->owner));
            /* this is for STABLE version */
#ifndef CS_AST_HAS_FLAG_MOH
			ast_clear_flag(CS_AST_BRIDGED_CHANNEL(c->owner), AST_FLAG_MOH);
#endif
	}
    else {
		ast_log(LOG_ERROR, "%s Can't resume the channel %s-%d. No bridged call\n", d->id, l->name, c->callid);
		return 0;
	}

	sccp_channel_set_active(c);
	sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONNECTED);
#ifdef CS_AST_CONTROL_HOLD
	ast_queue_control(c->owner, AST_CONTROL_UNHOLD);
#endif
	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Resume the channel %s-%d\n", d->id, l->name, c->callid);
	return 1;
}

void sccp_channel_delete(sccp_channel_t * c) {
	sccp_line_t * l;
	sccp_device_t * d;

	if (!c)
		return;

	ast_mutex_lock(&c->lock);

	l = c->line;
	d = c->device;
	
	if (d->transfer_channel == c)
		d->transfer_channel = NULL;

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Deleting channel %d from line %s\n", d->id, c->callid, c->line->name);

	/* mark the channel DOWN so any pending thread will terminate */
	if (c->owner) {
		ast_setstate(c->owner, AST_STATE_DOWN);
		c->owner = NULL;
	}

	/* this is in case we are destroying the session */
	if (c->state != SCCP_CHANNELSTATE_DOWN)
		sccp_indicate_nolock(c, SCCP_CHANNELSTATE_ONHOOK);

	if (c->rtp)
		sccp_channel_stop_rtp(c);

	/* remove from the global channels list */

	if (c->next) { /* not the last one */
		ast_mutex_lock(&c->next->lock);
		c->next->prev = c->prev;
		ast_mutex_unlock(&c->next->lock);
	}
	if (c->prev) { /* not the first one */
		ast_mutex_lock(&c->prev->lock);
		c->prev->next = c->next;
		ast_mutex_unlock(&c->prev->lock);
	}
	else { /* the first one */
		GLOB(channels) = c->next;
	}

	/* remove the channel from the channels line list */
	if (c->next_on_line) { /* not the last one */
		ast_mutex_lock(&c->next_on_line->lock);
		c->next_on_line->prev_on_line = c->prev_on_line;
		ast_mutex_unlock(&c->next_on_line->lock);
	}
	if (c->prev_on_line) { /* not the first one */
		ast_mutex_lock(&c->prev_on_line->lock);
		c->prev_on_line->next_on_line = c->next_on_line;
		ast_mutex_unlock(&c->prev_on_line->lock);
	}
	else { /* the first one */
		ast_mutex_lock(&l->lock);
		l->channels = c->next_on_line;
		ast_mutex_unlock(&l->lock);
	}

	ast_mutex_lock(&l->lock);
/*	if (l->activeChannel == c)
		l->activeChannel = NULL; */
	l->channelCount--;
	ast_mutex_unlock(&l->lock);

	/* deactive the active call if needed */
	ast_mutex_lock(&d->lock);
	if (d->active_channel == c)
		d->active_channel = NULL;
	ast_mutex_unlock(&d->lock);

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Deleted channel %d from line %s\n", c->device->id, c->callid, c->line->name);
	ast_mutex_unlock(&c->lock);
	free(c);
	return;
}

void sccp_channel_start_rtp(sccp_channel_t * c) {
	sccp_session_t * s;
	char iabuf[INET_ADDRSTRLEN];
	if (!c || !c->device)
		return;
	s = c->device->session;
	if (!s)
		return;
		
	ast_mutex_lock(&c->lock);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Creating rtp server connection at %s\n", c->device->id, ast_inet_ntoa(iabuf, sizeof(iabuf), s->ourip));
	c->rtp = ast_rtp_new_with_bindaddr(NULL, NULL, 1, 0, s->ourip);
//	ast_rtp_setnat(c->rtp, 0x1);

	if (c->rtp && c->owner)
		c->owner->fds[0] = ast_rtp_fd(c->rtp);

  ast_mutex_unlock(&c->lock);
}

void sccp_channel_stop_rtp(sccp_channel_t * c) {
	if (c->rtp) {
		if (c->owner)
			c->owner->fds[0] = -1;
		ast_rtp_destroy(c->rtp);
		c->rtp = NULL;
	}
}

void sccp_channel_transfer(sccp_channel_t * c) {
	sccp_device_t * d;
	if (!c)
		return;
		
	d = c->device;

	if (!d->transfer) {
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Tranfer disabled on device or line\n", d->id);
		return;
	}

	ast_mutex_lock(&d->lock);
	/* are we in the middle of a transfer? */
	if (d->transfer_channel && (d->transfer_channel != c)) {
		ast_mutex_unlock(&d->lock);
		sccp_channel_transfer_complete(c);
		return;
	}

	d->transfer_channel = c;
	ast_mutex_unlock(&d->lock);
	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Transfer request from line channel %s-%d\n", d->id, c->line->name, c->callid);

	if (!c->owner) {
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: No bridged channel to transfer on %s-%d\n", d->id, c->line->name, c->callid);
		sccp_dev_displayprompt(NULL, c, SKINNY_DISP_CAN_NOT_COMPLETE_TRANSFER, 5);
		return;
	}
	if ( (c->state != SCCP_CHANNELSTATE_HOLD && c->state != SCCP_CHANNELSTATE_CALLTRANSFER) && !sccp_channel_hold(c) )
		return;
	if (c->state != SCCP_CHANNELSTATE_CALLTRANSFER)
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_CALLTRANSFER);
	sccp_channel_newcall(c->line, NULL);
}

void sccp_channel_transfer_complete(sccp_channel_t * c) {
#ifndef CS_AST_CHANNEL_HAS_CID
	char *name, *number, *cidtmp;
#endif
	struct ast_channel	*transferred = NULL, *original_transferred=NULL,	*transferee = NULL, *destination = NULL;
	sccp_channel_t * peer;
	sccp_device_t * d = c->device;
	ast_mutex_lock(&d->lock);
	peer = d->transfer_channel;
	ast_mutex_unlock(&d->lock);
	
	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Complete transfer from %s-%d\n", d->id, c->line->name, c->callid);

	if (!c->owner || !peer->owner) {
			sccp_log(1)(VERBOSE_PREFIX_3 "%s: Transfer error, asterisk channel error %s-%d and %s-%d\n", d->id, c->line->name, c->callid, peer->line->name, peer->callid);
		return;
	}

	transferred = CS_AST_BRIDGED_CHANNEL(peer->owner);
	original_transferred = transferred;
	destination = CS_AST_BRIDGED_CHANNEL(c->owner);
	transferee = c->owner;

	if (!transferred || !transferee) {
		ast_log(LOG_WARNING, "Failed to complete transfer. Missing asterisk transferred or transferee channel\n");
		sccp_dev_displayprompt(NULL, c, SKINNY_DISP_CAN_NOT_COMPLETE_TRANSFER, 5);
		return;
	}
	
	if (transferee->cdr && transferred->cdr) {
		transferee->cdr = ast_cdr_append(transferee->cdr, transferred->cdr);
	} else if (transferred->cdr) {
		transferee->cdr = transferred->cdr;
	}
	transferred->cdr = NULL;

	if (ast_channel_masquerade(transferee, transferred)) {
		ast_log(LOG_WARNING, "Failed to masquerade %s into %s\n", transferee->name, transferred->name);
		sccp_dev_displayprompt(NULL, c, SKINNY_DISP_CAN_NOT_COMPLETE_TRANSFER, 5);
		return;
	}
	ast_mutex_unlock(&transferee->lock);
	if (d->transfer_channel);
		sccp_channel_endcall(d->transfer_channel);
	ast_mutex_lock(&d->lock);
	d->transfer_channel = NULL;
	ast_mutex_unlock(&d->lock);
	if (!destination || strncasecmp(destination->type,"SCCP",4))
		return;

	/* it's a SCCP channel destination on transfer */
	c = CS_AST_CHANNEL_PVT(destination);

	if (c) {
		sccp_dev_starttone(c->device, SKINNY_TONE_CALLWAITINGTONE);
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Transfer confirmation destination on channel %s\n", d->id, destination->name);
		/* display the transferred CID info to destination */
#ifdef CS_AST_CHANNEL_HAS_CID
		sccp_channel_set_callingparty(c, transferred->cid.cid_name, transferred->cid.cid_num);
#else
		if (transferred->callerid && (cidtmp = strdup(transferred->callerid))) {
			ast_callerid_parse(cidtmp, &name, &number);
			sccp_channel_set_callingparty(c, name, number);
			free(cidtmp);
		}
#endif
		sccp_channel_send_callinfo(c);
	}
}

#ifdef CS_SCCP_PARK

struct sccp_dual {
	struct ast_channel *chan1;
	struct ast_channel *chan2;
};

static void * sccp_channel_park_thread(void *stuff) {
	struct ast_channel *chan1, *chan2;
	struct sccp_dual *d;
	struct ast_frame *f;
	int ext;
	int res;
	char extstr[20];
	sccp_channel_t * c;
	memset(&extstr, 0 , sizeof(extstr));

	d = stuff;
	chan1 = d->chan1;
	chan2 = d->chan2;
	free(d);
	f = ast_read(chan1);
	if (f)
		ast_frfree(f);
	res = ast_park_call(chan1, chan2, 0, &ext);
	if (!res) {
		extstr[0] = 128;
		extstr[1] = SKINNY_LBL_CALL_PARK_AT;
		sprintf(&extstr[2],"%d",ext);
		c = CS_AST_CHANNEL_PVT(chan2);
		sccp_dev_displayprompt(NULL,c, extstr, 0);
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Parked channel %s on %d\n", c->device->id, chan1->name, ext);
	}
	ast_hangup(chan2);
	return NULL;
}

void sccp_channel_park(sccp_channel_t * c) {
	sccp_device_t * d;
	struct sccp_dual *dual;
	struct ast_channel *chan1m, *chan2m, *bridged;
	pthread_t th;
	
	if (!c)
		return;
	
	d = c->device;
	if (!d)
		return;
		
	if (!d->park) {
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Park disabled on device\n", d->id);
		return;
	}

	bridged = CS_AST_BRIDGED_CHANNEL(c->owner);
	if (!bridged || !c->owner) {
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Can't Park: no asterisk channel\n", d->id);
		return;
	}
	sccp_indicate_lock(c, SCCP_CHANNELSTATE_CALLPARK);
	
	chan1m = ast_channel_alloc(0);
	if (!chan1m) {
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Park Failed: can't create asterisk channel\n", d->id);
		sccp_dev_displayprompt(NULL,c, SKINNY_DISP_NO_PARK_NUMBER_AVAILABLE, 0);
		return;
	}
	chan2m = ast_channel_alloc(0);
	if (!chan1m) {
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Park Failed: can't create asterisk channel\n", d->id);
		sccp_dev_displayprompt(NULL,c, SKINNY_DISP_NO_PARK_NUMBER_AVAILABLE, 0);
		ast_hangup(chan1m);
		return;
	}
	
	snprintf(chan1m->name, sizeof(chan1m->name), "Parking/%s", bridged->name);
	/* Make formats okay */
	chan1m->readformat = bridged->readformat;
	chan1m->writeformat = bridged->writeformat;
	ast_channel_masquerade(chan1m, bridged);
	/* Setup the extensions and such */
	sccp_copy_string(chan1m->context, bridged->context, sizeof(chan1m->context));
	sccp_copy_string(chan1m->exten, bridged->exten, sizeof(chan1m->exten));
	chan1m->priority = bridged->priority;
	
	/* We make a clone of the peer channel too, so we can play
	   back the announcement */
	snprintf(chan2m->name, sizeof (chan2m->name), "SCCPParking/%s",c->owner->name);
	/* Make formats okay */
	chan2m->readformat = c->owner->readformat;
	chan2m->writeformat = c->owner->writeformat;
	ast_channel_masquerade(chan2m, c->owner);
	/* Setup the extensions and such */
	sccp_copy_string(chan2m->context, c->owner->context, sizeof(chan2m->context));
	sccp_copy_string(chan2m->exten, c->owner->exten, sizeof(chan2m->exten));
	chan2m->priority = c->owner->priority;
	if (ast_do_masquerade(chan2m)) {
		ast_log(LOG_WARNING, "SCCP: Masquerade failed :(\n");
		ast_hangup(chan2m);
		return;
	}

	dual = malloc(sizeof(struct sccp_dual));
	if (dual) {
		memset(d, 0, sizeof(*dual));
		dual->chan1 = chan1m;
		dual->chan2 = chan2m;
		if (!ast_pthread_create(&th, NULL, sccp_channel_park_thread, dual))
			return;
		free(dual);
	}
}
#endif
