/*
 * (SCCP*)
 *
 * An implementation of Skinny Client Control Protocol (SCCP)
 *
 * Sergio Chersovani (mlists@c-net.it)
 *
 * Reworked, but based on chan_sccp code.
 * The original chan_sccp driver that was made by Zozo which itself was derived from the chan_skinny driver.
 * Modified by Jan Czmok and Julien Goodwin
 *
 * This program is free software and may be modified and
 * distributed under the terms of the GNU Public License.
 */

#include "config.h"
#include "chan_sccp.h"
#include "sccp_actions.h"
#include "sccp_utils.h"
#include "sccp_device.h"
#include "sccp_channel.h"
#include "sccp_pbx.h"
#include "sccp_indicate.h"
#include <asterisk/callerid.h>
#include <asterisk/pbx.h>
#include <unistd.h>


static struct ast_frame * sccp_rtp_read(sccp_channel_t * c) {
	/* Retrieve audio/etc from channel.  Assumes c->lock is already held. */
	struct ast_frame * f;
	/* if c->rtp not null */
	if (!c->rtp)
		return NULL;

	f = ast_rtp_read(c->rtp);

	if (c->owner) {
		/* We already hold the channel lock */
		if (f->frametype == AST_FRAME_VOICE) {
			if (f->subclass != c->owner->nativeformats) {
				sccp_log(10)(VERBOSE_PREFIX_3 "%s: Oooh, format changed to %d on channel %d\n", c->device->id, f->subclass, c->callid);
				c->owner->nativeformats = f->subclass;
				ast_set_read_format(c->owner, c->owner->readformat);
				ast_set_write_format(c->owner, c->owner->writeformat);
			}
		}
	}
	return f;
}

/* this is for incoming calls asterisk sccp_request */
static int sccp_pbx_call(struct ast_channel *ast, char *dest, int timeout) {
  sccp_line_t	 * l;
  sccp_device_t  * d;
  sccp_session_t * s;
  sccp_channel_t * c;
  char * ringermode;

#ifndef CS_AST_CHANNEL_HAS_CID
  char * name, * number, *cidtmp; // For the callerid parse below
#endif
	c = CS_AST_CHANNEL_PVT(ast);

	if (!c) {
		ast_log(LOG_WARNING, "SCCP: Asterisk request to call %s channel: %s, but we don't have this channel!\n", dest, ast->name);
		return -1;
	}

	l = c->line;
	d = l->device;
	s = d->session;

	sccp_log(1)(VERBOSE_PREFIX_3 "%s: Asterisk request to call %s\n", d->id, ast->name);

	if (d->dnd) {
		ast_setstate(ast, AST_STATE_BUSY);
		ast_queue_control(ast, AST_CONTROL_BUSY);
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: DND is on. Call %s rejected\n", d->id, ast->name);
		return 0;
	}

	/* if incoming call limit is reached send BUSY */
	ast_mutex_lock(&l->lock);
	if ( l->channelCount > l->incominglimit ) { /* >= just to be sure :-) */
		sccp_log(1)(VERBOSE_PREFIX_3 "Incoming calls limit (%d) reached on SCCP/%s... sending busy\n", l->incominglimit, l->name);
		ast_mutex_unlock(&l->lock);
		ast_setstate(ast, AST_STATE_BUSY);
		ast_queue_control(ast, AST_CONTROL_BUSY);
		return 0;
	}
	ast_mutex_unlock(&l->lock);

  /* Set the channel callingParty Name and Number */
#ifdef CS_AST_CHANNEL_HAS_CID
	sccp_channel_set_callingparty(c, ast->cid.cid_name, ast->cid.cid_num);
#else
	if (ast->callerid && (cidtmp = strdup(ast->callerid))) {
		ast_callerid_parse(cidtmp, &name, &number);
		sccp_channel_set_callingparty(c, name, number);
		free(cidtmp);
	}
#endif

  /* Set the channel calledParty Name and Number 7910 compatibility*/
	sccp_channel_set_calledparty(c, l->cid_name, l->cid_num);

  /* find and set the ringer mode */
	ringermode = pbx_builtin_getvar_helper(ast, "ALERT_INFO");
	if ( ringermode && !ast_strlen_zero(ringermode) ) {
		sccp_log(1)(VERBOSE_PREFIX_3 "%s: Found ALERT_INFO=%s\n", d->id, ringermode);
		if (strcasecmp(ringermode, "inside") == 0)
			c->ringermode = SKINNY_STATION_INSIDERING;
		else if (strcasecmp(ringermode, "feature") == 0)
			c->ringermode = SKINNY_STATION_FEATURERING;
		else if (strcasecmp(ringermode, "silent") == 0)
			c->ringermode = SKINNY_STATION_SILENTRING;
		else if (strcasecmp(ringermode, "urgent") == 0)
			c->ringermode = SKINNY_STATION_URGENTRING;
		else
			c->ringermode = SKINNY_STATION_OUTSIDERING;
	}

	if ( sccp_channel_get_active(d) )
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_CALLWAITING);
	else
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_RINGIN);

	ast_queue_control(ast, AST_CONTROL_RINGING);
	return 0;
}


static int sccp_pbx_hangup(struct ast_channel * ast) {
	sccp_channel_t * c;
	sccp_line_t    * l;
	sccp_device_t  * d;

	c = CS_AST_CHANNEL_PVT(ast);

	ast_mutex_lock(&GLOB(usecnt_lock));
	GLOB(usecnt)--;
	ast_mutex_unlock(&GLOB(usecnt_lock));

	ast_update_use_count();

	if (!c) {
		sccp_log(10)(VERBOSE_PREFIX_3 "SCCP: Asked to hangup channel %s. SCCP channel already hangup\n", ast->name);
		return 0;
	}

	ast_mutex_lock(&GLOB(channels_lock));
	ast_mutex_lock(&c->lock);
	c->owner = NULL;
	l = c->line;
	d = l->device;

	sccp_log(1)(VERBOSE_PREFIX_3 "SCCP: Asterisk request to hangup %s channel %s\n", skinny_calltype2str(c->calltype), ast->name);

	CS_AST_CHANNEL_PVT(ast) = NULL;

	if (c->rtp) {
		sccp_channel_closereceivechannel(c);
		sccp_channel_stop_rtp(c);
	}

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Current channel %s-%d state %s(%d)\n", d->id, l->name, c->callid, sccp_indicate2str(c->state), c->state);

	if ( c->state != SCCP_CHANNELSTATE_DOWN) {
		if (c->state == SCCP_CHANNELSTATE_CONNECTED) {
			sccp_indicate_nolock(c, SCCP_CHANNELSTATE_HANGUP_CONNECTED);
		} else if (c->state == SCCP_CHANNELSTATE_RINGIN) {
			sccp_indicate_nolock(c, SCCP_CHANNELSTATE_HANGUP_RINGIN);
		} else if (c->state == SCCP_CHANNELSTATE_CALLWAITING) {
			sccp_indicate_nolock(c, SCCP_CHANNELSTATE_HANGUP_CALLWAITING);
		} else if (c->state == SCCP_CHANNELSTATE_HOLD || c->state == SCCP_CHANNELSTATE_CALLTRANSFER) {
			sccp_indicate_nolock(c, SCCP_CHANNELSTATE_HANGUP_HOLD);
		} else if (c->state == SCCP_CHANNELSTATE_DIALING) {
			sccp_indicate_nolock(c, SCCP_CHANNELSTATE_INVALIDNUMBER);
		} else if ( (c->state != SCCP_CHANNELSTATE_INVALIDNUMBER) && (c->state != SCCP_CHANNELSTATE_CONGESTION) ){
			sccp_indicate_nolock(c, SCCP_CHANNELSTATE_CONGESTION);
		}
	}

	if (c->state == SCCP_CHANNELSTATE_DOWN) {
		ast_mutex_unlock(&c->lock);
		sccp_channel_delete(c);
	} else
		ast_mutex_unlock(&c->lock);

	ast_mutex_unlock(&GLOB(channels_lock));
	sccp_channel_check_ringback(d);
	return 0;
}

static int sccp_pbx_answer(struct ast_channel *ast) {
	sccp_channel_t * c = CS_AST_CHANNEL_PVT(ast);

	if (!c || !c->device || !c->line) {
		ast_log(LOG_ERROR, "SCCP: Answered %s but no SCCP channel\n", ast->name);
		return -1;
	}

	sccp_log(1)(VERBOSE_PREFIX_3 "SCCP: Outgoing call has been answered %s on %s@%s-%d\n", ast->name, c->line->name, c->device->id, c->callid);
	sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONNECTED);
	return 0;
}


static struct ast_frame * sccp_pbx_read(struct ast_channel *ast) {
	sccp_channel_t * c = CS_AST_CHANNEL_PVT(ast);
	struct ast_frame *fr;
	ast_mutex_lock(&c->lock);
	fr = sccp_rtp_read(c);
	ast_mutex_unlock(&c->lock);
	return fr;
}

static int sccp_pbx_write(struct ast_channel *ast, struct ast_frame *frame) {
	sccp_channel_t * c = CS_AST_CHANNEL_PVT(ast);

	int res = 0;
	if (frame->frametype != AST_FRAME_VOICE) {
		if (frame->frametype == AST_FRAME_IMAGE)
			return 0;
		else {
			ast_log(LOG_WARNING, "%s: Can't send %d type frames with SCCP write on channel %d\n", (c) ? c->device->id : "SCCP", frame->frametype, (c) ? c->callid : 0);
			return 0;
		}
	} else {
		if (!(frame->subclass & ast->nativeformats)) {
			ast_log(LOG_WARNING, "%s: Asked to transmit frame type %d, while native formats is %d (read/write = %d/%d)\n",
			(c) ? c->device->id : "SCCP", frame->subclass, ast->nativeformats, ast->readformat, ast->writeformat);
			return -1;
		}
	}
	if (c) {
		ast_mutex_lock(&c->lock);
		if (c->rtp) {
			res =  ast_rtp_write(c->rtp, frame);
		}
		ast_mutex_unlock(&c->lock);
	}
	return res;
}

static char *sccp_control2str(int state) {
		switch(state) {
		case AST_CONTROL_HANGUP:
				return "Hangup";
		case AST_CONTROL_RING:
				return "Ring";
		case AST_CONTROL_RINGING:
				return "Ringing";
		case AST_CONTROL_ANSWER:
				return "Answer";
		case AST_CONTROL_BUSY:
				return "Busy";
		case AST_CONTROL_TAKEOFFHOOK:
				return "TakeOffHook";
		case AST_CONTROL_OFFHOOK:
				return "OffHook";
		case AST_CONTROL_CONGESTION:
				return "Congestion";
		case AST_CONTROL_FLASH:
				return "Flash";
		case AST_CONTROL_WINK:
				return "Wink";
		case AST_CONTROL_OPTION:
				return "Option";
		case AST_CONTROL_RADIO_KEY:
				return "RadioKey";
		case AST_CONTROL_RADIO_UNKEY:
				return "RadioUnKey";
		case AST_CONTROL_PROGRESS:
				return "Progress";
		case AST_CONTROL_PROCEEDING:
				return "Proceeding";
#ifdef CS_AST_CONTROL_HOLD
		case AST_CONTROL_HOLD:
				return "Hold";
		case AST_CONTROL_UNHOLD:
				return "UnHold";
#endif
		default:
				return "Unknown";
		}
}

static int sccp_pbx_indicate(struct ast_channel *ast, int ind) {
	sccp_channel_t * c = CS_AST_CHANNEL_PVT(ast);
	if (!c)
		return -1;

	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Asterisk indicate '%d' (%s) condition on channel %s\n", (c) ? c->device->id : "SCCP", ind, sccp_control2str(ind), ast->name);
	if (c->state == SCCP_CHANNELSTATE_PROCEED) {
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: we are in PROCEED state on channel %s... skip the indication\n", (c) ? c->device->id : "SCCP", ast->name);
		return 0;
	}

	switch(ind) {

	case AST_CONTROL_RINGING:
		/* workaround for iax providers that send a ringout after answer indication */
		if (c->state == SCCP_CHANNELSTATE_CONNECTED)
			return -1; /* let's asterisk emulate it */
		sccp_indicate_lock(c,SCCP_CHANNELSTATE_RINGOUT);
		break;

	case AST_CONTROL_BUSY:
		sccp_indicate_lock(c,SCCP_CHANNELSTATE_BUSY);
		break;

	case AST_CONTROL_CONGESTION:
		sccp_indicate_lock(c,SCCP_CHANNELSTATE_CONGESTION);
		break;

	case AST_CONTROL_PROGRESS:
	case AST_CONTROL_PROCEEDING:
		sccp_indicate_lock(c,SCCP_CHANNELSTATE_PROCEED);
		break;

#ifdef CS_AST_CONTROL_HOLD
/* when the bridged channel hold/unhold the call we are notified here */
	case AST_CONTROL_HOLD:
		break;
	case AST_CONTROL_UNHOLD:
		break;
#endif

	case -1:
		break;

	default:
	  ast_log(LOG_WARNING, "SCCP: Don't know how to indicate condition %d\n", ind);
	  return -1;
  }
  return 0;
}

static int sccp_pbx_fixup(struct ast_channel *oldchan, struct ast_channel *newchan) {
	sccp_channel_t * c = CS_AST_CHANNEL_PVT(newchan);

	ast_mutex_lock(&c->lock);
	ast_log(LOG_NOTICE, "sccp_pbx_fixup(%s, %s)\n", oldchan->name, newchan->name);
	if (c->owner != oldchan) {
		ast_log(LOG_WARNING, "old channel wasn't %p but was %p\n", oldchan, c->owner);
		ast_mutex_unlock(&c->lock);
		return -1;
	}
	c->owner = newchan;
	ast_mutex_unlock(&c->lock);
	return 0;
}

static int sccp_pbx_recvdigit(struct ast_channel *ast, char digit) {
	sccp_channel_t * c = CS_AST_CHANNEL_PVT(ast);

	if (digit == '*') {
		digit = 0xe; // See the definition of tone_list in chan_sccp.h for more info
	} else if (digit == '#') {
		digit = 0xf;
	} else if (digit == '0') {
		digit = 0xa; /* 0 is not 0 for cisco :-) Cisco goes Ace high */
	} else {
		digit -= '0';
	}

	if (digit > 16)
		return -1;

	sccp_log(1)(VERBOSE_PREFIX_3 "Asked to send tone '%d'\n", digit);
	sccp_dev_starttone(c->device, (uint8_t) digit);
	return 0;
}

#ifdef CS_AST_HAS_TECH_PVT
const struct ast_channel_tech sccp_tech = {
	.type = "SCCP",
	.description = "Skinny Client Control Protocol (SCCP)",
	.capabilities = AST_FORMAT_ALAW|AST_FORMAT_ULAW|AST_FORMAT_G729A,
	.requester = sccp_request,
/*	  .devicestate = sccp_devicestate, */
	.call = sccp_pbx_call,
	.hangup = sccp_pbx_hangup,
	.answer = sccp_pbx_answer,
	.read = sccp_pbx_read,
	.write = sccp_pbx_write,
	.indicate = sccp_pbx_indicate,
	.fixup = sccp_pbx_fixup,
	.send_digit = sccp_pbx_recvdigit,
	/* .bridge = ast_rtp_bridge */
};
#endif


uint8_t sccp_pbx_channel_allocate(sccp_channel_t * c) {
	sccp_device_t * d = c->device;
	struct ast_channel * tmp;
	sccp_line_t * l = c->line;
	int fmt;
#ifndef CS_AST_CHANNEL_HAS_CID
	char cidtmp[256];
	memset(&cidtmp, 0, sizeof(cidtmp));
#endif

	tmp = ast_channel_alloc(1);
	if (!tmp) {
		ast_log(LOG_ERROR, "%s: Unable to allocate asterisk channel on line %s\n", d->id, l->name);
		return 0;
	}

	/* need to reset the exten, otherwise it would be set to s */
	memset(&tmp->exten,0,sizeof(tmp->exten));

	/* let's connect the ast channel to the sccp channel */
	ast_mutex_lock(&c->lock);
	c->owner = tmp;
	ast_mutex_unlock(&c->lock);
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Global Capabilities: %d\n", d->id, GLOB(global_capability));

	ast_mutex_lock(&l->lock);
	ast_mutex_lock(&d->lock);
	if (d->capability) {
	  tmp->nativeformats = ast_codec_choose(&d->codecs, d->capability, 1);
	} else {
	  tmp->nativeformats = ast_codec_choose(&GLOB(global_codecs), GLOB(global_capability), 1);
	}
	ast_mutex_unlock(&l->lock);
	ast_mutex_unlock(&d->lock);

	fmt = ast_best_codec(tmp->nativeformats);

	if (GLOB(debug) > 2) {
	  const unsigned slen=512;
	  char s1[slen];
	  sccp_log(2)(VERBOSE_PREFIX_3 "%s: Channel %s, capabilities: DEVICE %s NATIVE %s BEST %s\n",
		d->id,
		c->owner->name,
		ast_getformatname_multiple(s1, slen, d->capability),
		ast_getformatname_multiple(s1, slen, tmp->nativeformats),
		ast_getformatname(fmt));
	}

	snprintf(tmp->name, sizeof(tmp->name), "SCCP/%s-%010x", l->name, c->callid);

	tmp->type = "SCCP";
	tmp->writeformat		 = fmt;
	tmp->readformat 		 = fmt;

#ifdef CS_AST_HAS_TECH_PVT
	tmp->tech				 = &sccp_tech;
	tmp->tech_pvt			 = c;
#else
	tmp->pvt->pvt			 = c;
	tmp->pvt->rawwriteformat = fmt;
	tmp->pvt->rawreadformat  = fmt;

	tmp->pvt->answer		 = sccp_pbx_answer;
	tmp->pvt->hangup		 = sccp_pbx_hangup;
	tmp->pvt->call			 = sccp_pbx_call;

	tmp->pvt->read			 = sccp_pbx_read;
	tmp->pvt->write 		 = sccp_pbx_write;
/*	  tmp->pvt->devicestate    = sccp_devicestate; */
	tmp->pvt->indicate		 = sccp_pbx_indicate;
	tmp->pvt->fixup 		 = sccp_pbx_fixup;
	tmp->pvt->send_digit	 = sccp_pbx_recvdigit;
#endif

	tmp->adsicpe		 = AST_ADSI_UNAVAILABLE;

	// XXX: Bridge?
	// XXX: Transfer?

	ast_mutex_lock(&GLOB(usecnt_lock));
	GLOB(usecnt)++;
	ast_mutex_unlock(&GLOB(usecnt_lock));

	ast_update_use_count();

#ifdef CS_AST_CHANNEL_HAS_CID
	if (l->cid_num)
	  tmp->cid.cid_num = strdup(l->cid_num);
	if (l->cid_name)
	  tmp->cid.cid_name = strdup(l->cid_name);
#else
	snprintf(cidtmp, sizeof(cidtmp), "\"%s\" <%s>", l->cid_name, l->cid_num);
	tmp->callerid = strdup(cidtmp);
#endif

	sccp_copy_string(tmp->context, l->context, sizeof(tmp->context));

	tmp->priority = 1;
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: Allocated asterisk channel %s-%d\n", d->id, l->name, c->callid);
	return 1;
}

void * sccp_pbx_startchannel(void *data) {
	struct ast_channel * chan = data;
	sccp_channel_t * c;
	sccp_line_t * l;
	sccp_device_t * d;
	uint8_t res_exten = 0, res_wait = 0, res_timeout = 0;

	c = CS_AST_CHANNEL_PVT(chan);
	if ( !c || !(l = c->line) || !(d = c->device) ) {
		ast_hangup(chan);
		return NULL;
	}

	sccp_log(1)( VERBOSE_PREFIX_3 "%s: New call on line %s\n", d->id, l->name);

	c->calltype = SKINNY_CALLTYPE_OUTBOUND;

	sccp_channel_set_callingparty(c, l->cid_name, l->cid_num);

	if (!ast_strlen_zero(chan->exten)) {
		/* we have a number to dial. Let's do it */
		sccp_log(10)( VERBOSE_PREFIX_3 "%s: Dialing %s on channel %s-%d\n", l->device->id, chan->exten, l->name, c->callid);
		sccp_copy_string(d->lastNumber, chan->exten, sizeof(d->lastNumber));
		sccp_channel_set_calledparty(c, chan->exten, chan->exten);
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_DIALING);
		/* Answer dialplan command works only when in RINGING OR RING ast_state */
		sccp_ast_setstate(c, AST_STATE_RING);
		if (ast_pbx_run(chan)) {
			if (d->state != SCCP_DEVICESTATE_ONHOOK)
				sccp_indicate_lock(c, SCCP_CHANNELSTATE_CONGESTION);
		}
		return NULL;
	}

	/* we have to collect the number */
	/* the phone is on TsOffHook state */
	sccp_log(10)( VERBOSE_PREFIX_3 "%s: Waiting for the number to dial on channel %s-%d\n", l->device->id, l->name, c->callid);
	/* let's use the keypad to collect digits */
	ast_mutex_lock(&c->lock);
	c->digittimeout = time(0)+GLOB(firstdigittimeout);
	ast_mutex_unlock(&c->lock);

	res_exten = 1;

	do {
		if (!ast_strlen_zero(chan->exten)) {
			res_exten = ast_matchmore_extension(chan, chan->context, chan->exten, 1, l->cid_num);
		}
		/* looking for a control event AST_STATE_DOWN */
		ast_safe_sleep(chan, 500);
		pthread_testcancel();
		if (! (res_wait = ( (chan->_state == AST_STATE_DOWN) || chan->_softhangup )) ) {
			if (CS_AST_CHANNEL_PVT(chan)) {
				ast_mutex_lock(&c->lock);
				res_timeout = (time(0) < c->digittimeout);
				ast_mutex_unlock(&c->lock);
			} else
				res_timeout = 0;
		}
	} while ( (res_wait == 0) && res_exten &&  res_timeout);

	if (res_wait != 0) {
		/* AST_STATE_DOWN or softhangup */
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: return from the startchannel for DOWN cause\n", l->device->id);
		if (chan->_softhangup)
			ast_hangup(chan);
		return NULL;
	}

	if ( !ast_strlen_zero(chan->exten)
			&& ast_exists_extension(chan, chan->context, chan->exten, 1, l->cid_num) ) {
		/* found an extension, let's dial it */
		sccp_log(10)(VERBOSE_PREFIX_3 "%s: channel %s-%d is dialing number %s\n", l->device->id, l->name, c->callid, chan->exten);
		sccp_copy_string(d->lastNumber, chan->exten, sizeof(d->lastNumber));
		sccp_channel_set_calledparty(c, chan->exten, chan->exten);
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_DIALING);
		/* Answer dialplan command works only when in RINGING OR RING ast_state */
		sccp_ast_setstate(c, AST_STATE_RING);
		if (ast_pbx_run(chan)) {
			sccp_indicate_lock(c, SCCP_CHANNELSTATE_INVALIDNUMBER);
		}
		return NULL;
	} else {
		/* timeout and no extension match */
		sccp_indicate_lock(c, SCCP_CHANNELSTATE_INVALIDNUMBER);
		ast_hangup(chan);
	}
	return NULL;
}

void sccp_pbx_senddigit(sccp_channel_t * c, char digit) {
	struct ast_frame f = { AST_FRAME_DTMF, };

	f.subclass = digit;

	ast_mutex_lock(&c->lock);
	ast_queue_frame(c->owner, &f);
	ast_mutex_unlock(&c->lock);
}

void sccp_pbx_senddigits(sccp_channel_t * c, char digits[AST_MAX_EXTENSION]) {
	int i;
	struct ast_frame f = { AST_FRAME_DTMF, };

	// We don't just call sccp_pbx_senddigit due to potential overhead, and issues with locking
	ast_mutex_lock(&c->lock);
	for (i = 0; digits[i] != '\0'; i++) {
		f.subclass = digits[i];
		ast_queue_frame(c->owner, &f);
	}
	ast_mutex_unlock(&c->lock);
}
