/*
 * (SCCP*)
 *
 * An implementation of Skinny Client Control Protocol (SCCP)
 *
 * Sergio Chersovani (mlists@c-net.it)
 *
 * Reworked, but based on chan_sccp code.
 * The original chan_sccp driver that was made by Zozo which itself was derived from the chan_skinny driver.
 * Modified by Jan Czmok and Julien Goodwin
 *
 * This program is free software and may be modified and
 * distributed under the terms of the GNU Public License.
 */

#include "chan_sccp.h"
#include "sccp_actions.h"
#include "sccp_utils.h"
#include "sccp_pbx.h"
#include "sccp_channel.h"
#include "sccp_device.h"
#include "sccp_line.h"
#include "sccp_indicate.h"

void sccp_sk_redial(sccp_device_t * d , sccp_line_t * l, sccp_channel_t * c) {
    sccp_log(10)(VERBOSE_PREFIX_3 "%s: Redial Softkey.\n",d->id);

    if (!d->lastNumber) {
        sccp_log(10)(VERBOSE_PREFIX_3 "%s: No number to redial\n", d->id);
        return;
	}

	if (c) {
		if (c->state == SCCP_CHANNELSTATE_OFFHOOK) {
			/* we have a offhook channel */
			ast_mutex_lock(&c->lock);
			sccp_copy_string(c->owner->exten, d->lastNumber, sizeof(c->owner->exten));
			sccp_log(10)(VERBOSE_PREFIX_3 "%s: Get ready to redial number %s\n", d->id, c->owner->exten);
			c->digittimeout = time(0)+1;
			ast_mutex_unlock(&c->lock);
		}
		/* here's a KEYMODE error. nothing to do */
		return;
	}
    if (!l)
        l = d->currentLine;
    sccp_channel_newcall(l, d->lastNumber);
}

void sccp_sk_newcall(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    if (!l)
        l = d->currentLine;
	sccp_channel_newcall(l, NULL);
}

void sccp_sk_hold(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    if (!c) {
        sccp_log(10)(VERBOSE_PREFIX_3 "### Hold Softkey while no call -> IGNORED\n");
        sccp_dev_displayprompt(d,c, "No call to put on hold.",5);
        return;
    }
    sccp_log(10)(VERBOSE_PREFIX_3 "### Hold Softkey pressed\n");
    sccp_channel_hold(c);
}

void sccp_sk_resume(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    // This is simply the reverse of the Hold Function.
    if (!c) {
         sccp_log(10)(VERBOSE_PREFIX_3 "### ResumeCall Softkey but no Call present - Ignoring \n");
         return;
	}

	sccp_log(10)(VERBOSE_PREFIX_3 "### ResumeCall Softkey pressed\n");
	sccp_channel_resume(c);
}

void sccp_sk_transfer(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
	if (!c) {
	    sccp_log(10)(VERBOSE_PREFIX_3 "### Transfer when there is no active call\n");
	    return;
	}
	
	sccp_channel_transfer(c);
	
}

void sccp_sk_endcall(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    if (!c) {
	    sccp_log(10)(VERBOSE_PREFIX_3 "%s: Endcall with no call in progress\n", d->id);
        return;
    }
    sccp_channel_endcall(c);
}


void sccp_sk_dnd(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
	/* change this to a per-line DND */

    ast_mutex_lock(&d->lock);

    d->dnd = (d->dnd) ? 0 : 1;

    if (d->dnd) {
        sccp_dev_displayprompt(d, NULL, ">>> " SKINNY_DISP_DND " <<<", 0);
    } else {
        sccp_dev_clearprompt(d, NULL);
        sccp_dev_displayprompt(d, NULL, SKINNY_DISP_YOUR_CURRENT_OPTIONS, 0);
    }
    ast_mutex_unlock(&d->lock);
    sccp_dev_dbput(d);
}


void sccp_sk_backspace(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
	int len;
	if (!c)
		return;
	if (c->state != SCCP_CHANNELSTATE_DIALING)
		return;
	ast_mutex_lock(&c->lock);
	len = strlen(c->owner->exten)-1;
	if (len >= 0)
    	c->owner->exten[len] = '\0';
	sccp_log(10)(VERBOSE_PREFIX_3 "%s: backspacing dial number %s\n", c->device->id, c->owner->exten);
	ast_mutex_unlock(&c->lock);
}

void sccp_sk_answer(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
	sccp_channel_answer(c);
}

void sccp_sk_dirtrfr(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
/*	sccp_channel_dirtrfr(c); */
}

void sccp_sk_cfwdall(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    sccp_log(10)(VERBOSE_PREFIX_3 "### CFwdAll Softkey pressed - NOT SUPPORTED\n");
}

void sccp_sk_cfwdbusy(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    sccp_log(10)(VERBOSE_PREFIX_3 "### CFwdBusy Softkey pressed - NOT SUPPORTED\n");
}

void sccp_sk_cfwdnoanswer(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    sccp_log(10)(VERBOSE_PREFIX_3 "### CFwdNoAnswer Softkey pressed - NOT SUPPORTED\n");
}

void sccp_sk_park(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
#ifdef CS_SCCP_PARK
    sccp_channel_park(c);
#else
    sccp_log(10)(VERBOSE_PREFIX_3 "### Native park was not compiled in\n");
#endif
}

void sccp_sk_trnsfvm(sccp_device_t * d, sccp_line_t * l, sccp_channel_t * c) {
    sccp_log(10)(VERBOSE_PREFIX_3 "### TransfVM Softkey pressed - NOT SUPPORTED\n");
}
